"""Parser for OpenDRIVE (.xodr) files."""

import math
import xml.etree.ElementTree as ET
import numpy as np
from scipy.integrate import quad
from scipy.integrate import odeint
from pynverse import inversefunc
from shapely.geometry import Polygon, MultiPolygon, GeometryCollection, Point, MultiPoint
from shapely.ops import unary_union, snap
import abc

# Lane types on which cars can appear.
DRIVABLE = [
    'driving',
    'entry',
    'exit',
    'offRamp',
    'onRamp'
]

# Lane types representing sidewalks.
SIDEWALK = ['sidewalk']

def buffer_union(polygons, buf=0.1):
    return unary_union([p.buffer(buf) for p in polygons]).buffer(-buf)

class Poly3:
    '''Cubic polynomial.'''
    def __init__(self, a, b, c, d):
        self.a = a
        self.b = b
        self.c = c
        self.d = d

    def eval_at(self, x):
        return self.a + self.b * x + self.c * x ** 2 + self.d * x ** 3

    def grad_at(self, x):
        return self.b + 2 * self.c * x + 3 * self.d * x ** 2


class Curve:
    ''' Geometric elements which compose road reference lines.
    See the OpenDRIVE Format Specification for coordinate system details.'''
    def __init__(self, x0, y0, hdg, length):
        self.x0 = x0
        self.y0 = y0
        self.hdg = hdg    # In radians counterclockwise, 0 at positive x-axis.
        self.length = length

    @abc.abstractmethod
    def to_points(self, num):
        '''Sample NUM evenly-spaced points from curve.
        Points are tuples of (x, y, s) with (x, y) absolute coordinates
        and s the arc length along the curve.'''
        return

    def rel_to_abs(self, points):
        '''Convert from relative coordinates of curve to absolute coordinates.
        I.e. rotate counterclockwise by self.hdg and translate by (x0, x1).'''
        def translate(p):
            return (p[0] + self.x0, p[1] + self.y0, p[2])

        def rotate(p):
            return (np.cos(self.hdg) * p[0] - np.sin(self.hdg) * p[1],
                    np.sin(self.hdg) * p[0] + np.cos(self.hdg) * p[1],
                    p[2])

        return [translate(rotate(p)) for p in points]


class Cubic(Curve):
    '''A curve defined by the cubic polynomial a + bu + cu^2 + du^3.
    The curve starts at (X0, Y0) in direction HDG, with length LENGTH.'''
    def __init__(self, x0, y0, hdg, length, a, b, c, d):
        super().__init__(x0, y0, hdg, length)
        self.poly = Poly3(a, b, c, d)

    def to_points(self, num):
        def arclength(s):
            d_arc = lambda x: np.sqrt(1 + self.poly.grad_at(x) ** 2)
            return quad(d_arc, 0, s)[0]

        s1 = inversefunc(arclength, self.length).tolist()
        points = [(s, self.poly.eval_at(s), arclength(s)) for s in np.linspace(0, s1, num=num)]
        return self.rel_to_abs(points)


class ParamCubic(Curve):
    ''' A curve defined by the parametric equations
    u = a_u + b_up + c_up^2 + d_up^3,
    v = a_v + b_vp + c_vp^2 + d_up^3,
    with p in [0, p_range].
    The curve starts at (X0, Y0) in direction HDG, with length LENGTH.'''
    def __init__(self, x0, y0, hdg, length,
                 au, bu, cu, du, av, bv, cv, dv, p_range=1):
        super().__init__(x0, y0, hdg, length)
        self.u_poly = Poly3(au, bu, cu, du)
        self.v_poly = Poly3(av, bv, cv, dv)
        self.p_range = p_range if p_range else 1

    def to_points(self, num):
        def eval_poly(s):
            return (self.u_poly.eval_at(s), self.v_poly.eval_at(s))

        def arclength(s):
            d_arc = lambda x: np.sqrt(self.u_poly.grad_at(s) ** 2
                                      + self.v_poly.grad_at(s))
            return quad(d_arc, 0, s)[0]

        points = [eval_poly(s) + (arclength(s),) for s in np.linspace(0, self.p_range, num=num)]
        return self.rel_to_abs(points)


class Clothoid(Curve):
    '''An Euler spiral with curvature varying linearly between CURV0 and CURV1.
    The spiral starts at (X0, Y0) in direction HDG, with length LENGTH.'''
    def __init__(self, x0, y0, hdg, length, curv0, curv1):
        super().__init__(x0, y0, hdg, length)
        # Initial and final curvature.
        self.curv0 = curv0
        self.curv1 = curv1

    def to_points(self, num):
        # Generate a origin-centered clothoid with zero curvature at origin,
        # then translate/rotate the relevant segment.
        # Arcs are just a degenerate clothiod:
        if self.curv0 == self.curv1:
            r = 1 / abs(self.curv0)
            theta = self.length / r
            th_space = np.linspace(0, theta, num=num)
            if self.curv0 == 0:
                points = [(x, 0, x) for x in np.linspace(0, self.length, num=num)]
            elif self.curv0 > 0:
                points = [(r * np.sin(th), r - r * np.cos(th), r * th) for th in th_space]
            else:
                points = [(r * np.sin(th), -r + r * np.cos(th), r * th) for th in th_space]
            return self.rel_to_abs(points)

        else:
            curve_rate = (self.curv1 - self.curv0) / self.length
            def clothoid_ode(state, s):
                x, y, theta = state[0], state[1], state[2]
                return np.array([np.cos(theta), np.sin(theta), self.curv0 + curve_rate * s])

            s_space = np.linspace(0, self.length, num)
            points = odeint(clothoid_ode, np.array([self.x0,self.y0,self.hdg]), s_space)
            return [(points[i,0], points[i,1], s_space[i]) for i in range(len(s_space))]


class Line(Curve):
    '''A line segment between (x0, y0) and (x1, y1).'''
    def __init__(self, x0, y0, hdg, length):
        super().__init__(x0, y0, hdg, length)
        # Endpoints of line.
        self.x1 = x0 + length * math.cos(hdg)
        self.y1 = y0 + length * math.sin(hdg)

    def to_points(self, num=10):
        x = np.linspace(self.x0, self.x1, num=num)
        y = np.linspace(self.y0, self.y1, num=num)
        s = [np.sqrt((x[i] - self.x0) **2 + (y[i] - self.y0) ** 2)
             for i in range(num)]
        return  list(zip(x, y, s))


class Lane():
    def __init__(self, id_, type_, pred=None, succ=None):
        self.id_ = id_
        self.width = []    # List of tuples (Poly3, int) for width and s-offset.
        self.type_ = type_
        self.pred = pred
        self.succ = succ

    def width_at(self, s):
        # S here is relative to start of LaneSection this lane is in.
        ind = 0
        while ind + 1 < len(self.width) and self.width[ind + 1][1] <= s:
            ind += 1
        assert self.width[ind][1] <= s, 'No matching width entry found.'
        w_poly, s_off = self.width[ind]
        assert(w_poly.eval_at(s - s_off) >= 0), 'Negative width!'
        return w_poly.eval_at(s - s_off)


class LaneSection():
    def __init__(self, s0, left_lanes={}, right_lanes={}):
        self.s0 = s0
        self.left_lanes = left_lanes
        self.right_lanes = right_lanes
        self.lanes = dict(list(left_lanes.items()) + list(right_lanes.items()))

    def get_lane(self, id_):
        if id_ in self.left_lanes:
            return self.left_lanes[id_]
        elif id_ in self.right_lanes:
            return self.right_lanes[id_]
        elif id_ == 0:
            return Lane(0, 'none')
        else:
            raise RuntimeError('Lane with id', id_, 'not found')

    def get_offsets(self, s):
        '''Returns dict of lane id and offset from
        reference line of lane boundary at coordinate S along line.
        By convention, left lanes have positive width offset and right lanes
        have negative.'''
        assert s >= self.s0, 'Input s is before lane start position.'
        offsets = {}
        left_lane_ids = sorted(self.left_lanes.keys())
        right_lane_ids = sorted(self.right_lanes.keys(), reverse=True)
        for lane_id in left_lane_ids:
            if lane_id - 1 in left_lane_ids:
                offsets[lane_id] = offsets[lane_id - 1] \
                    + self.left_lanes[lane_id].width_at(s - self.s0)
            else:
                offsets[lane_id] = self.left_lanes[lane_id].width_at(s - self.s0)
        for lane_id in right_lane_ids:
            if lane_id + 1 in right_lane_ids:
                offsets[lane_id] = offsets[lane_id + 1] \
                    - self.right_lanes[lane_id].width_at(s - self.s0)
            else:
                offsets[lane_id] = -self.right_lanes[lane_id].width_at(s - self.s0)
        return offsets


class RoadLink:
    '''Indicates Roads a and b, with ids id_a and id_b respectively, are connected.'''
    def __init__(self, id_a, id_b, contact_a, contact_b):
        self.id_a = id_a
        self.id_b = id_b
        # contact_a and contact_b should be of value "start" or "end"
        # and indicate which end of each road is connected to the other.
        self.contact_a = contact_a
        self.contact_b = contact_b


class Junction:
    class Connection:
        def __init__(self, incoming_id, connecting_id, connecting_contact):
            self.incoming_id = incoming_id
            # id of connecting road
            self.connecting_id = connecting_id
            # contact point ('start' or 'end') on connecting road
            self.connecting_contact = connecting_contact

    def __init__(self, id_):
        self.id_ = id_
        self.connections = []
        # Ids of roads that are paths within junction:
        self.paths = []

    def add_connection(self, incoming_id, connecting_id, connecting_contact):
        self.connections.append(Junction.Connection(incoming_id,
                                                    connecting_id,
                                                    connecting_contact))


class Road:
    def __init__(self, name, id_, length, drive_on_right=True):
        self.name = name
        self.id_ = id_
        self.length = length
        self.lane_secs = []    # List of LaneSection objects.
        self.ref_line = []    # List of Curve objects defining reference line.
        # NOTE: sec_points, sec_polys, sec_lane_polys should be ordered according to lane_secs.
        self.sec_points = []   # List of lists of points, one for each LaneSection.
        self.sec_polys = []   # List of Polygons, one for each LaneSections.
        self.sec_lane_polys = []    # List of dict of lane id to Polygon for each LaneSection.
        self.lane_polys = []    # List of lane polygons. Not a dict b/c lane id is not unique along road.
        # Each polygon in lane_polys is the union of connected lane section polygons.
        # lane_polys is currently not used.
        # Reference line offset:
        self.offset = []    # List of tuple (Poly3, s-coordinate).
        self.drive_on_right = drive_on_right
        # Used to fill in gaps between roads:
        self.start_bounds_left = {}
        self.start_bounds_right = {}
        self.end_bounds_left = {}
        self.end_bounds_right = {}

    def get_lane(self, id_, s):
        '''Returns Lane object with id_ at coordinate S along line.'''
        ind = 0
        while ind + 1 < len(self.lane_secs) and self.lane_secs[ind + 1].s0 <= s:
            ind += 1
        assert self.lane_secs[ind].s0 <= s, 'No matching lane section found.'
        return self.lane_secs[ind].get_lane(id_)

    def get_ref_line_offset(self, s):
        if not self.offset:
            return 0
        ind = 0
        while ind + 1 < len(self.offset) and self.offset[ind + 1][1] < s:
            ind += 1
        return self.offset[ind][0].eval_at(s)

    def get_ref_points(self, num):
        '''Returns list of list of points for each piece of ref_line.
        List of list structure necessary because each piece needs to be
        constructed into Polygon separately then unioned afterwards to avoid
        self-intersecting lines.'''
        ref_points = []
        for piece in self.ref_line:
            piece_points = piece.to_points(num)
            assert piece_points, 'Failed to get piece points'
            if ref_points:
                piece_points = [(p[0], p[1], p[2] + ref_points[-1][-1][2])
                                for p in piece_points]
            ref_points.append(piece_points)
        return ref_points

    def get_lane_offsets(self, s):
        '''Returns dict of lane id and offset from
        reference line of lane boundary at coordinate S along line.'''
        s = float(s)
        ind = 0
        while ind + 1 < len(self.lane_secs) and self.lane_secs[ind + 1].s0 <= s:
            ind += 1
        assert self.lane_secs[ind].s0 <= s, 'No matching lane section found.'
        offsets = self.lane_secs[ind].get_offsets(s)
        offsets[0] = 0    # Center lane has width 0 by convention.
        for id_ in offsets.keys():
            offsets[id_] += self.get_ref_line_offset(s)
        return offsets

    def heading_at(self, point):
        # Convert point to shapely Point.
        point = Point(point.x, point.y)
        for i in range(len(self.lane_secs)):
            ref_points = self.sec_points[i]
            poly = self.sec_polys[i]
            if point.within(poly.buffer(1)):
                lane_id = None
                for id_ in self.sec_lane_polys[i].keys():
                    if point.within(self.sec_lane_polys[i][id_].buffer(1)):
                        lane_id = id_
                        break
                assert lane_id is not None, 'Point not found in sec_lane_polys.'
                min_dist = float('inf')
                for i in range(len(ref_points)):
                    cur_point = Point(ref_points[i][0], ref_points[i][1])
                    if point.distance(cur_point) < min_dist:
                        closest_idx = i
                if closest_idx >= len(ref_points) - 1:
                   closest_idx = len(ref_points) - 2
                dy = ref_points[closest_idx + 1][1] - ref_points[closest_idx][1]
                dx = ref_points[closest_idx + 1][0] - ref_points[closest_idx][0]
                heading = math.atan2(dy, dx)
                # Right lanes have negative lane_id.
                # Flip heading if drive_on_right XOR right lane.
                if self.drive_on_right != (lane_id < 0):
                    heading += math.pi
                # Heading 0 is defined differently between OpenDrive and Scenic(?)
                heading -= math.pi / 2
                return (heading + math.pi) % (2 * math.pi) - math.pi

        raise RuntimeError('Point not found in piece_polys')

    def calc_geometry_for_type(self, lane_types, num, calc_gap=False):
        '''Given a list of lane types, returns a tuple of:
        - List of lists of points along the reference line, with same indexing as self.lane_secs
        - List of region polygons, with same indexing as self.lane_secs
        - List of dictionary of lane id to polygon, with same indexing as self.lane_secs
        - List of polygons for each lane (not necessarily by id, but respecting lane successor/predecessor)
        - Polygon for entire region.
        If calc_gap=True, fills in gaps between connected roads. This is fairly expensive.'''
        road_polygons = []
        ref_points = self.get_ref_points(num)
        cur_lane_polys = {}
        sec_points = []
        sec_polys = []
        sec_lane_polys = []
        lane_polys = []
        last_lefts = None
        last_rights = None

        for i in range(len(self.lane_secs)):
            cur_sec = self.lane_secs[i]
            cur_sec_points = []
            if i < len(self.lane_secs) - 1:
                next_sec = self.lane_secs[i + 1]
                s_stop = next_sec.s0
            else:
                s_stop = float('inf')
            left_bounds = {}
            right_bounds = {}
            cur_sec_lane_polys = {}
            cur_sec_polys = []
            # Last point in left/right lane boundary line for last road piece:
            start_of_sec = True
            end_of_sec = False

            while ref_points and not end_of_sec:
                if not ref_points[0] or ref_points[0][0][2] >= s_stop:
                    # Case 1: The current list of ref_points (corresponding to current piece)
                    # is empty, so we move onto the next list of points.
                    # Case 2: The s-coordinate has exceeded s_stop, so we should move
                    # onto the next LaneSection.
                    # Either way, we collect all the bound points so far into polygons.
                    if not ref_points[0]:
                        ref_points.pop(0)
                    else:
                       end_of_sec = True
                    cur_last_lefts = {}
                    cur_last_rights = {}
                    for id_ in left_bounds.keys():
                        # Polygon for piece of lane:
                        bounds = left_bounds[id_] + right_bounds[id_][::-1]
                        if len(bounds) < 3:
                            continue
                        poly = Polygon(bounds).buffer(0)
                        if poly.is_valid and not poly.is_empty:
                            if poly.geom_type == 'MultiPolygon':
                                poly = MultiPolygon([p for p in list(poly)
                                                     if not p.is_empty and p.exterior])
                                cur_sec_polys.extend(list(poly))
                            else:
                                cur_sec_polys.append(poly)
                            if id_ in cur_sec_lane_polys:
                                cur_sec_lane_polys[id_].append(poly)
                            else:
                                cur_sec_lane_polys[id_] = [poly]
                        if calc_gap:
                            # Polygon for gap between lanes:
                            if start_of_sec:
                                prev_id = cur_sec.lanes[id_].pred
                            else:
                                prev_id = id_
                            if last_lefts is not None and prev_id in last_lefts.keys():
                                gap_poly = MultiPoint([
                                    last_lefts[prev_id], last_rights[prev_id],
                                    left_bounds[id_][0], right_bounds[id_][0]]).convex_hull
                                assert gap_poly.is_valid, 'Gap polygon not valid.'
                                # Assume MultiPolygon cannot result from convex hull.
                                if gap_poly.geom_type == 'Polygon' and not gap_poly.is_empty:
                                    # plot_poly(gap_poly, 'b')
                                    gap_poly = gap_poly.buffer(0)
                                    cur_sec_polys.append(gap_poly)
                                    if id_ in cur_sec_lane_polys:
                                        cur_sec_lane_polys[id_].append(gap_poly)
                                    else:
                                        cur_sec_lane_polys[id_] = [gap_poly]
                        cur_last_lefts[id_] = left_bounds[id_][-1]
                        cur_last_rights[id_] = right_bounds[id_][-1]
                        if (start_of_sec and i == 0) or not self.start_bounds_left:
                            self.start_bounds_left[id_] = left_bounds[id_][0]
                            self.start_bounds_right[id_] = right_bounds[id_][0]

                    left_bounds = {}
                    right_bounds = {}
                    if cur_last_lefts and cur_last_rights:
                        last_lefts = cur_last_lefts
                        last_rights = cur_last_rights
                        start_of_sec = False
                else:
                    cur_p = ref_points[0].pop(0)
                    cur_sec_points.append(cur_p)
                    offsets = cur_sec.get_offsets(cur_p[2])
                    offsets[0] = 0
                    for id_ in offsets.keys():
                        offsets[id_] += self.get_ref_line_offset(cur_p[2])
                    if ref_points[0]:
                        next_p = ref_points[0][0]
                        tan_vec = (next_p[0] - cur_p[0],
                                   next_p[1] - cur_p[1])
                    else:
                        if len(cur_sec_points) >= 2:
                            prev_p = cur_sec_points[-2]
                        else:
                            assert len(sec_points) > 0
                            if sec_points[-1]:
                                prev_p = sec_points[-1][-1]
                            else:
                                prev_p = sec_points[-2][-1]

                        tan_vec = (cur_p[0] - prev_p[0],
                                   cur_p[1] - prev_p[1])
                    tan_norm = np.sqrt(tan_vec[0] ** 2 + tan_vec[1] ** 2)
                    # if tan_norm < 0.01:
                    #     continue
                    normal_vec = (-tan_vec[1] / tan_norm, tan_vec[0] / tan_norm)
                    for id_ in offsets.keys():
                        if cur_sec.get_lane(id_).type_ in lane_types:
                            if id_ > 0:
                                prev_id = id_ - 1
                            else:
                                prev_id = id_ + 1
                            left_bound = [cur_p[0] + normal_vec[0] * offsets[id_],
                                          cur_p[1] + normal_vec[1] * offsets[id_]]
                            right_bound = [cur_p[0] + normal_vec[0] * offsets[prev_id],
                                           cur_p[1] + normal_vec[1] * offsets[prev_id]]
                            if id_ not in left_bounds:
                                left_bounds[id_] = [left_bound]
                            else:
                                left_bounds[id_].append(left_bound)
                            if id_ not in right_bounds:
                                right_bounds[id_] = [right_bound]
                            else:
                                right_bounds[id_].append(right_bound)
            sec_points.append(cur_sec_points)
            sec_polys.append(buffer_union(cur_sec_polys))
            for id_ in cur_sec_lane_polys:
                cur_sec_lane_polys[id_] = buffer_union(cur_sec_lane_polys[id_])
            sec_lane_polys.append(cur_sec_lane_polys)
            next_lane_polys = {}
            for id_ in cur_sec_lane_polys:
                pred_id = cur_sec.get_lane(id_).pred
                if pred_id and pred_id in cur_lane_polys:
                    next_lane_polys[id_] = cur_lane_polys.pop(pred_id) \
                        + [cur_sec_lane_polys[id_]]
                else:
                    next_lane_polys[id_] = [cur_sec_lane_polys[id_]]
            for id_ in cur_lane_polys:
                lane_polys.append(buffer_union(cur_lane_polys[id_]))
            cur_lane_polys = next_lane_polys
        for id_ in cur_lane_polys:
            lane_polys.append(buffer_union(cur_lane_polys[id_]))
        union_poly = buffer_union(sec_polys)
        if last_lefts and last_rights:
            self.end_bounds_left.update(last_lefts)
            self.end_bounds_right.update(last_rights)
        return sec_points, sec_polys, sec_lane_polys, lane_polys, union_poly

    def calculate_geometry(self, num, calc_gap=False):
        # Note: this also calculates self.start_bounds_left, ,self.start_bounds_right,
        # self.end_bounds_left, self.end_bounds_right
        self.sec_points,\
            self.sec_polys,\
            self.sec_lane_polys,\
            self.lane_polys,\
            self.drivable_region =\
                self.calc_geometry_for_type(DRIVABLE, num, calc_gap=calc_gap)
        _, _, _, _, self.sidewalk_region = self.calc_geometry_for_type(SIDEWALK, num, calc_gap=calc_gap)

class RoadMap:
    def __init__(self):
        self.roads = {}
        self.road_links = []
        self.junctions = {}
        self.sec_lane_polys = []
        self.lane_polys = []
        self.intersection_region = None

    def calculate_geometry(self, num, calc_gap=False, calc_intersect=False):
        # If calc_gap=True, fills in gaps between connected roads.
        # If calc_intersect=True, calculates intersection regions.
        # These are fairly expensive.
        for road in self.roads.values():
            road.calculate_geometry(num, calc_gap=calc_gap)
        drivable_polys = {}
        sidewalk_polys = {}
        drivable_gap_polys = []
        sidewalk_gap_polys = []
        for road in self.roads.values():
            self.sec_lane_polys.extend(road.sec_lane_polys)
            self.lane_polys.extend(road.lane_polys)
            drivable_poly = road.drivable_region
            sidewalk_poly = road.sidewalk_region
            if not (drivable_poly is None or drivable_poly.is_empty):
                drivable_polys[road.id_] = drivable_poly.buffer(0.001)
            if not (sidewalk_poly is None or sidewalk_poly.is_empty):
                sidewalk_polys[road.id_] = sidewalk_poly.buffer(0.001)

        for link in self.road_links:
            road_a = self.roads[link.id_a]
            road_b = self.roads[link.id_b]
            assert link.contact_a in ['start', 'end'], 'Invalid link record.'
            assert link.contact_b in ['start', 'end'], 'Invalid link record.'
            if link.contact_a == 'start':
                a_sec = road_a.lane_secs[0]
                a_bounds_left = road_a.start_bounds_left
                a_bounds_right = road_a.start_bounds_right
            else:
                a_sec = road_a.lane_secs[-1]
                a_bounds_left = road_a.end_bounds_left
                a_bounds_right = road_a.end_bounds_right
            if link.contact_b == 'start':
                b_bounds_left = road_b.start_bounds_left
                b_bounds_right = road_b.start_bounds_right
            else:
                b_bounds_left = road_b.end_bounds_left
                b_bounds_right = road_b.end_bounds_right
            for id_, lane in a_sec.lanes.items():
                if link.contact_a == 'start':
                    other_id = lane.pred
                else:
                    other_id = lane.succ
                if other_id not in b_bounds_left or other_id not in b_bounds_right:
                    continue
                if id_ not in a_bounds_left or id_ not in a_bounds_right:
                    continue
                if calc_gap:
                    gap_poly = MultiPoint([
                        a_bounds_left[id_], a_bounds_right[id_],
                        b_bounds_left[other_id], b_bounds_right[other_id]
                    ]).convex_hull
                    if not gap_poly.is_valid:
                        continue
                    # assert gap_poly.is_valid, 'Gap polygon not valid.'
                    if gap_poly.geom_type == 'Polygon' and not gap_poly.is_empty:
                        if lane.type_ in DRIVABLE:
                            drivable_gap_polys.append(gap_poly)
                            # plot_poly(gap_poly, 'g')
                        elif lane.type_ in SIDEWALK:
                            sidewalk_gap_polys.append(gap_poly)
                            # plot_poly(gap_poly, 'k')

        self.drivable_region = buffer_union(list(drivable_polys.values()) + drivable_gap_polys)
        self.sidewalk_region = buffer_union(list(sidewalk_polys.values()) + sidewalk_gap_polys)
        self.calculate_intersections()

    def calculate_intersections(self):
        intersect_polys = []
        for junc in self.junctions.values():
            junc_roads = set()
            for conn in junc.connections:
               junc_roads.add(conn.incoming_id)
               junc_roads.add(conn.connecting_id)
            for road_i_idx in junc_roads:
                if road_i_idx in junc.paths:
                    road_i = self.roads[road_i_idx]
                    intersect_polys.append(road_i.drivable_region)
                    continue
                for road_j_idx in junc_roads:
                    if road_i_idx == road_j_idx:
                        continue
                    road_i = self.roads[road_i_idx]
                    road_j = self.roads[road_j_idx]
                    intersect_polys.append(
                        road_i.drivable_region.intersection(road_j.drivable_region))
        self.intersection_region = buffer_union(intersect_polys)

    def heading_at(self, point):
        '''Return the road heading at point.'''
        # Convert point to shapely Point.
        point = Point(point.x, point.y)
        for road in self.roads.values():
            if point.within(road.drivable_region.buffer(1)):
                return road.heading_at(point)
        #raise RuntimeError('Point not in RoadMap: ', point)
        return 0

    def plot_line(self, plt, num=500):
        '''Plot center line of road map for sanity check.'''
        for road in self.roads.values():
            for piece in road.ref_line:
                points = piece.to_points(num)
                x = [p[0] for p in points]
                y = [p[1] for p in points]
                plt.plot(x, y, 'b')
        plt.show()

    def plot_lanes(self, plt, num=500):
        '''Plot lane boundaries of road map for sanity check.'''
        bounds_x =[]
        bounds_y = []
        for road in self.roads.values():
            for piece in road.ref_line:
                ref_points = piece.to_points(num)
                for i in range(len(ref_points) - 1):
                    offsets = road.get_lane_offsets(ref_points[i][2])
                    tan_vec = (ref_points[i + 1][0] - ref_points[i][0],
                               ref_points[i + 1][1] - ref_points[i][1])
                    tan_norm = np.sqrt(tan_vec[0] ** 2 + tan_vec[1] ** 2)
                    normal_vec = (-tan_vec[1] / tan_norm, tan_vec[0] / tan_norm)
                    # ortho_line_x = []
                    # ortho_line_y = []
                    for id_ in offsets.keys():
                        if road.get_lane(id_, ref_points[i][2]).type_ == 'driving':
                            bounds_x.append(ref_points[i][0] + normal_vec[0] * offsets[id_])
                            bounds_y.append(ref_points[i][1] + normal_vec[1] * offsets[id_])
        plt.scatter(bounds_x, bounds_y, c='r', s=2)
        plt.show()

    def __parse_lanes(self, lanes_elem):
        '''Lanes_elem should be <left> or <right> element.
        Returns dict of lane ids and Lane objects.'''
        lanes = {}
        for l in lanes_elem.iter('lane'):
            id_ = int(l.get('id'))
            type_ = l.get('type')
            link = l.find('link')
            pred = None
            succ = None
            if link is not None:
                pred_elem = link.find('predecessor')
                succ_elem = link.find('successor')
                if pred_elem is not None:
                    pred = int(pred_elem.get('id'))
                if succ_elem is not None:
                    succ = int(succ_elem.get('id'))
            lane = Lane(id_, type_, pred, succ)
            for w in l.iter('width'):
                w_poly = Poly3(float(w.get('a')),
                               float(w.get('b')),
                               float(w.get('c')),
                               float(w.get('d')))
                lane.width.append((w_poly, float(w.get('sOffset'))))
            lanes[id_] = lane
        return lanes

    def __parse_link(self, link_elem, road_id, contact):
        if link_elem is None:
            return
        if link_elem.get('elementType') == 'road':
            self.road_links.append(RoadLink(road_id,
                                            int(link_elem.get('elementId')),
                                            contact,
                                            link_elem.get('contactPoint')))
        else:
            assert link_elem.get('elementType') == 'junction', 'Unknown link type'
            junction = int(link_elem.get('elementId'))
            connections = self.junctions[junction].connections
            for c in connections:
                if c.incoming_id == road_id:
                    self.road_links.append(RoadLink(road_id,
                                                    c.connecting_id,
                                                    contact,
                                                    c.connecting_contact))


    def parse(self, path):
        tree = ET.parse(path)
        root = tree.getroot()
        if root.tag != 'OpenDRIVE':
            raise RuntimeError(f'{path} does not appear to be an OpenDRIVE file')

        for j in root.iter('junction'):
            junction = Junction(int(j.get('id')))
            for c in j.iter('connection'):
                junction.add_connection(int(c.get('incomingRoad')),
                                        int(c.get('connectingRoad')),
                                        c.get('contactPoint'))
                junction.paths.append(int(c.get('connectingRoad')))
            self.junctions[junction.id_] = junction

        for r in root.iter('road'):
            link = r.find('link')
            if link is not None:
                pred_elem = link.find('predecessor')
                succ_elem = link.find('successor')
                self.__parse_link(pred_elem, int(r.get('id')), 'start')
                self.__parse_link(succ_elem, int(r.get('id')), 'end')
            road = Road(r.get('name'), int(r.get('id')), float(r.get('length')))

            # Parse planView:
            plan_view = r.find('planView')
            for geom in plan_view.iter('geometry'):
                x0 = float(geom.get('x'))
                y0 = float(geom.get('y'))
                s0 = float(geom.get('s'))
                hdg = float(geom.get('hdg'))
                length = float(geom.get('length'))
                curve_elem = geom[0]
                curve = None
                if curve_elem.tag == 'line':
                    curve = Line(x0, y0, hdg, length)
                elif curve_elem.tag == 'arc':
                    # Arc is clothoid of constant curvature.
                    curv = float(curve_elem.get('curvature'))
                    curve = Clothoid(x0, y0, hdg, length, curv, curv)
                elif curve_elem.tag == 'spiral':
                    curv0 = float(curve_elem.get('curvStart'))
                    curv1 = float(curve_elem.get('curvEnd'))
                    curve = Clothoid(x0, y0, hdg, length, curv0, curv1)
                elif curve_elem.tag == 'poly3':
                    a, b, c, d = cubic_elem.get('a'), \
                        float(curve_elem.get('b')), \
                        float(curve_elem.get('c')), \
                        float(curve_elem.get('d'))
                    curve = Cubic(x0, y0, hdg, length, a, b, c, d)
                elif curve_elem.tag == 'paramPoly3':
                    au, bu, cu, du, av, bv, cv, dv, p_range = \
                        float(curve_elem.get('aU')), \
                        float(curve_elem.get('bU')), \
                        float(curve_elem.get('cU')), \
                        float(curve_elem.get('dU')), \
                        float(curve_elem.get('aV')), \
                        float(curve_elem.get('bV')), \
                        float(curve_elem.get('cV')), \
                        float(curve_elem.get('dV')), \
                        float(curve_elem.get('pRange'))
                    curve = ParamCubic(x0, y0, hdg, length,
                                       au, bu, cu, du, av, bv,
                                       cv, dv, p_range)
                road.ref_line.append(curve)

            # Parse lanes:
            lanes = r.find('lanes')
            for offset in lanes.iter('laneOffset'):
                road.offset.append((Poly3(float(offset.get('a')),
                                         float(offset.get('b')),
                                         float(offset.get('c')),
                                         float(offset.get('d'))),
                                   float(offset.get('s'))))

            for ls_elem in lanes.iter('laneSection'):
                s = ls_elem.get('s')
                left = ls_elem.find('left')
                right = ls_elem.find('right')
                left_lanes = {}
                right_lanes = {}

                if left is not None:
                    left_lanes = self.__parse_lanes(left)

                if right is not None:
                    right_lanes = self.__parse_lanes(right)

                lane_sec = LaneSection(float(ls_elem.get('s')), left_lanes, right_lanes)
                road.lane_secs.append(lane_sec)
            self.roads[road.id_] = road




"""Extracting relations (for later pruning) from the syntax of requirements."""

import math
from collections import defaultdict
from ast import Compare, BinOp, Eq, NotEq, Lt, LtE, Gt, GtE, Call, Add, Sub, Expression, Name

from scenic.core.distributions import needsSampling
from scenic.core.object_types import Point, Object
from scenic.core.utils import InvalidScenarioError, InconsistentScenarioError

def inferRelationsFrom(reqNode, namespace, ego, line):
    """Infer relations between objects implied by a requirement."""
    matcher = RequirementMatcher(namespace)

    inferRelativeHeadingRelations(matcher, reqNode, ego, line)
    inferDistanceRelations(matcher, reqNode, ego, line)

def inferRelativeHeadingRelations(matcher, reqNode, ego, line):
    """Infer bounds on relative headings from a requirement."""
    rhMatcher = lambda node: matcher.matchUnaryFunction('RelativeHeading', node)
    allBounds = matcher.matchBounds(reqNode, rhMatcher)
    for target, bounds in allBounds.items():
        if not isinstance(target, Object):
            continue
        assert target is not ego
        if ego is None:
            raise InvalidScenarioError('relative heading w.r.t. unassigned ego on line {line}')
        lower, upper = bounds
        if lower < -math.pi:
            lower = -math.pi
        if upper > math.pi:
            upper = math.pi
        if lower == -math.pi and upper == math.pi:
            continue    # skip trivial bounds
        rel = RelativeHeadingRelation(target, lower, upper)
        ego._relations.append(rel)
        conv = RelativeHeadingRelation(ego, -upper, -lower)
        target._relations.append(conv)

def inferDistanceRelations(matcher, reqNode, ego, line):
    """Infer bounds on distances from a requirement."""
    distMatcher = lambda node: matcher.matchUnaryFunction('DistanceFrom', node)
    allBounds = matcher.matchBounds(reqNode, distMatcher)
    for target, bounds in allBounds.items():
        if not isinstance(target, Object):
            continue
        assert target is not ego
        if ego is None:
            raise InvalidScenarioError('distance w.r.t. unassigned ego on line {line}')
        lower, upper = bounds
        if lower < 0:
            lower = 0
            if upper == float('inf'):
                continue    # skip trivial bounds
        rel = DistanceRelation(target, lower, upper)
        ego._relations.append(rel)
        conv = DistanceRelation(ego, lower, upper)
        target._relations.append(conv)

class BoundRelation:
    """Abstract relation bounding something about another object."""
    def __init__(self, target, lower, upper):
        self.target = target
        self.lower, self.upper = lower, upper

class RelativeHeadingRelation(BoundRelation):
    """Relation bounding another object's relative heading with respect to this one."""
    pass

class DistanceRelation(BoundRelation):
    """Relation bounding another object's distance from this one."""
    pass

class RequirementMatcher:
    def __init__(self, namespace):
        self.namespace = namespace

    def inconsistencyError(self, node, message):
        raise InconsistentScenarioError(node.lineno, message)

    def matchUnaryFunction(self, name, node):
        """Match a call to a specified unary function, returning the value of its argument."""
        if not (isinstance(node, Call) and isinstance(node.func, Name)
                and node.func.id == name):
            return None
        if len(node.args) != 1:
            return None
        if len(node.keywords) != 0:
            return None
        return self.matchValue(node.args[0])

    def matchBounds(self, node, matchAtom):
        """Match upper/lower bounds on something matched by the given function.

        Returns a dict of all bounds found, mapping the bounded quantity to a
        pair (low, high) of lower/upper bounds.
        """
        if not isinstance(node, Compare):
            return {}
        bounds = defaultdict(lambda: (float('-inf'), float('inf')))
        first = node.left
        for second, op in zip(node.comparators, node.ops):
            lower, upper, target = self.matchBoundsInner(first, second, op, matchAtom)
            first = second
            if target is None:
                continue
            bestLower, bestUpper = bounds[target]
            if lower is not None and lower > bestLower:
                bestLower = lower
            if upper is not None and upper < bestUpper:
                bestUpper = upper
            bounds[target] = (bestLower, bestUpper)
        return bounds

    def matchBoundsInner(self, left, right, op, matchAtom):
        """Extract bounds from a single comparison operator."""
        # Reduce > and >= to < and <=
        if isinstance(op, Gt):
            return self.matchBoundsInner(right, left, Lt(), matchAtom)
        elif isinstance(op, GtE):
            return self.matchBoundsInner(right, left, LtE(), matchAtom)
        # Try matching a constant lower bound on the atom or its absolute value
        lconst = self.matchConstant(left)
        if isinstance(lconst, (int, float)):
            target = matchAtom(right)
            if target is not None:     # CONST op QUANTITY
                return (lconst, lconst, target) if isinstance(op, Eq) else (lconst, None, target)
            else:
                bounds = self.matchAbsBounds(right, lconst, op, False, matchAtom)
                if bounds is not None:      # CONST op abs(QUANTITY [+/- CONST])
                    return bounds
        # Try matching a constant upper bound on the atom or its absolute value
        rconst = self.matchConstant(right)
        if isinstance(rconst, (int, float)):
            target = matchAtom(left)
            if target is not None:      # QUANTITY op CONST
                return (rconst, rconst, target) if isinstance(op, Eq) else (None, rconst, target)
            else:
                bounds = self.matchAbsBounds(left, rconst, op, True, matchAtom)
                if bounds is not None:      # abs(QUANTITY [+/- CONST]) op CONST
                    return bounds
        return None, None, None

    def matchAbsBounds(self, node, const, op, isUpperBound, matchAtom):
        """Extract bounds on an atom from a comparison involving its absolute value."""
        if not (isinstance(node, Call) and isinstance(node.func, Name)
                and node.func.id == 'abs'):
            return None     # not an invocation of abs
        if not isUpperBound and not isinstance(op, Eq):
            return None     # lower bounds on abs value don't bound underlying quantity
        if const < 0:
            self.inconsistencyError(node, f'absolute value cannot be negative')
        assert len(node.args) == 1
        arg = node.args[0]
        target = matchAtom(arg)
        if target is not None:   # abs(QUANTITY) </= CONST
            return (-const, const, target)
        elif isinstance(arg, BinOp) and isinstance(arg.op, (Add, Sub)):   # abs(X +/- Y) </= CONST
            match = None
            slconst = self.matchConstant(arg.left)
            target = matchAtom(arg.right)
            if (isinstance(slconst, (int, float))
                and target is not None):   # abs(CONST +/- QUANTITY) </= CONST
                match = slconst
            else:
                srconst = self.matchConstant(arg.right)
                target = matchAtom(arg.left)
                if (isinstance(srconst, (int, float))
                    and target is not None):    # abs(QUANTITY +/- CONST) </= CONST
                    match = srconst
            if match is not None:
                if isinstance(arg.op, Add):    # abs(QUANTITY + CONST) </= CONST
                    return (-const - match, const - match, target)
                else:   # abs(QUANTITY - CONST) </= CONST
                    return (-const + match, const + match, target)
        return None

    def matchConstant(self, node):
        """Match constant values, i.e. values known prior to sampling."""
        value = self.matchValue(node)
        return None if needsSampling(value) else value

    def matchValue(self, node):
        """Match any expression which can be evaluated, returning its value.

        This method could have undesirable side-effects, but conditions in
        requirements should not have side-effects to begin with.
        """
        try:
            code = compile(Expression(node), '<internal>', 'eval')
            value = eval(code, dict(self.namespace))
        except Exception:
            return None
        return value



"""Support for values which are sampled outside of Scenic.

External Samplers in General
============================

External samplers provide a mechanism to use different types of sampling
techniques, like optimization or quasi-random sampling, from within a Scenic
program. Ordinary random values in Scenic are instances of `Distribution`;
this module defines a special subclass, `ExternalParameter`, representing a
value which is sampled externally. Scenic programs with external parameters
are handled as follows:

	1. During compilation, all instances of `ExternalParameter` are gathered
	   together and given to the `ExternalSampler.forParameters` function;
	   this function creates an appropriate `ExternalSampler`,
	   whose configuration can be controlled using various global parameters
	   (``param`` statements).

	2. When sampling a scene, before sampling any other distributions the
	   :obj:`~ExternalSampler.sample` method of the `ExternalSampler` is
	   called to sample all the external parameters. For active samplers, this
	   method passes along the ``feedback`` value given to `Scenario.generate`,
	   if any.

	3. Once the external parameters have values, the program is equivalent to
	   one without external parameters, and sampling proceeds as usual. As for
	   every instance of `Distribution`, the external parameters will have
	   their :obj:`~Samplable.sampleGiven` method called once all their
	   dependencies have been sampled; by default this method just returns the
	   value sampled for this parameter in step (2).

.. note::

	Note that while external parameters, like all instances of `Distribution`,
	are allowed to have dependencies, they are an exception to the usual rule
	that dependencies are always sampled before dependents, because the
	`ExternalSampler.sample` method is called before any other sampling.
	However, as explained above, the :obj:`~Samplable.sampleGiven` method is
	called in the proper order and external samplers which need to do sampling
	based on the values of other distributions can be invoked from it. The
	two-step mechanism with `ExternalSampler.sample` is provided for samplers
	which sample the whole space of external parameters at once (e.g. the
	VerifAI samplers).

Samplers from VerifAI
=====================

The external sampling mechanism is designed to be extensible. The only built-in
`ExternalSampler` is the `VerifaiSampler`, which provides access to the
samplers in the `VerifAI`_ toolkit (which in turn can use Scenic as a modeling
language).

The `VerifaiSampler` supports several types of external parameters corresponding
to the primitive distributions: `VerifaiRange` and `VerifaiDiscreteRange` for
continuous and discrete intervals, and `VerifaiOptions` for discrete sets.
For example, suppose we write::

	ego = Object at VerifaiRange(5, 15) @ 0

This is equivalent to the ordinary Scenic line ``ego = Object at (5, 15) @ 0``,
except that the X coordinate of the ego is sampled by VerifAI within the range
(5, 15) instead of being uniformly distributed over it. By default the
`VerifaiSampler` uses VerifAI's `Halton`_ sampler, so the range will still be
covered uniformly but more systematically. If we want to use a different sampler,
we can set the ``verifaiSamplerType`` global parameter::

	param verifaiSamplerType = 'ce'
	ego = Object at VerifaiRange(5, 15) @ 0

Now the X coordinate will be sampled using VerifAI's `cross-entropy`_ sampler.
If we pass a feedback value to `Scenario.generate` which scores the previous
scene, then the coordinate will not be sampled uniformly but rather converge to
a distribution concentrated on values minimizing the score. Active samplers like
cross-entropy can be used for falsification in this way, driving a system toward
parts of the parameter space where a specification is violated.

The cross-entropy sampler in VerifAI can be started from a non-uniform prior.
Scenic provides a convenient way to define this prior using the ordinary syntax
for distributions::

	param verifaiSamplerType = 'ce'
	ego = Object at VerifaiParameter.withPrior(Normal(10, 3)) @ 0

Now cross-entropy sampling will start from a normal distribution with mean 10
and standard deviation 3. Priors are restricted to primitive distributions and
in general may be approximated so that VerifAI can handle them -- see
`VerifaiParameter.withPrior` for details.

For more information on how to customize the sampler, see `VerifaiSampler`.

.. _VerifAI: https://github.com/BerkeleyLearnVerify/VerifAI

.. _Halton: https://en.wikipedia.org/wiki/Halton_sequence

.. _cross-entropy: https://en.wikipedia.org/wiki/Cross-entropy_method

"""

from dotmap import DotMap
import numpy

from scenic.core.distributions import Distribution, Options
from scenic.core.utils import InvalidScenarioError

class ExternalSampler:
	"""Abstract class for objects called to sample values for each external parameter.

	Attributes:
		rejectionFeedback: Value passed to the `sample` method when the last sample was rejected.
		  This value can be chosen by a Scenic scenario using the global parameter
		  ``externalSamplerRejectionFeedback``.
	"""
	def __init__(self, params, globalParams):
		# feedback value passed to external sampler when the last scene was rejected
		self.rejectionFeedback = globalParams.get('externalSamplerRejectionFeedback')

	@staticmethod
	def forParameters(params, globalParams):
		"""Create an `ExternalSampler` given the sets of external and global parameters.

		The scenario may explicitly select an external sampler by assigning the global
		parameter ``externalSampler`` to a subclass of `ExternalSampler`. Otherwise, a
		`VerifaiSampler` is used by default.

		Args:
			params (tuple): Tuple listing each `ExternalParameter`.
			globalParams (dict): Dictionary of global parameters for the `Scenario`.
			  Note that the values of these parameters may be instances of `Distribution`!

		Returns:
			An `ExternalSampler` configured for the given parameters.
		"""
		if len(params) > 0:
			externalSampler = globalParams.get('externalSampler', VerifaiSampler)
			if not issubclass(externalSampler, ExternalSampler):
				raise InvalidScenarioError(f'externalSampler type {externalSampler}'
				                           ' not subclass of ExternalSampler')
			return externalSampler(params, globalParams)
		else:
			return None

	def sample(self, feedback):
		"""Sample values for all the external parameters.

		Args:
			feedback: Feedback from the last sample (for active samplers).
		"""
		self.cachedSample = self.nextSample(feedback)

	def nextSample(self, feedback):
		"""Actually do the sampling. Implemented by subclasses."""
		raise NotImplementedError

	def valueFor(self, param):
		"""Return the sampled value for a parameter. Implemented by subclasses."""
		raise NotImplementedError

class VerifaiSampler(ExternalSampler):
	"""An external sampler exposing the samplers in the VerifAI toolkit.

	The sampler can be configured using the following Scenic global parameters:

		* ``verifaiSamplerType`` -- sampler type (see the ``verifai.server.choose_sampler``
		  function); the default is ``'halton'``
		* ``verifaiSamplerParams`` -- ``DotMap`` of options passed to the sampler

	The `VerifaiSampler` supports external parameters which are instances of `VerifaiParameter`.
	"""

	def __init__(self, params, globalParams):
		super().__init__(params, globalParams)
		import verifai.features
		import verifai.server

		# construct FeatureSpace
		usingProbs = False
		self.params = tuple(params)
		for index, param in enumerate(self.params):
			if not isinstance(param, VerifaiParameter):
				raise RuntimeError(f'VerifaiSampler given parameter of wrong type: {param}')
			param.sampler = self
			param.index = index
			if param.probs is not None:
				usingProbs = True
		space = verifai.features.FeatureSpace({
		    f'param{index}': verifai.features.Feature(param.domain)
		    for index, param in enumerate(self.params)
		})

		# set up VerifAI sampler
		samplerType = globalParams.get('verifaiSamplerType', 'halton')
		samplerParams = globalParams.get('verifaiSamplerParams', None)
		if usingProbs and samplerType == 'ce':
			if samplerParams is None:
				samplerParams = DotMap()
			if 'cont' in samplerParams or 'disc' in samplerParams:
				raise RuntimeError('CE distributions specified in both VerifaiParameters'
				                   'and verifaiSamplerParams')
			cont_buckets = []
			cont_dists = []
			disc_dists = []
			for param in self.params:
				if isinstance(param, VerifaiRange):
					if param.probs is None:
						buckets = 5
						dist = numpy.ones(buckets) / buckets
					else:
						dist = numpy.array(param.probs)
						buckets = len(dist)
					cont_buckets.append(buckets)
					cont_dists.append(dist)
				elif isinstance(param, VerifaiDiscreteRange):
					n = param.high - param.low + 1
					dist = numpy.ones(n)/n if param.probs is None else numpy.array(param.probs)
					disc_dists.append(dist)
				else:
					raise RuntimeError(f'Parameter {param} not supported by CE sampler')
			samplerParams.cont.buckets = cont_buckets
			samplerParams.cont.dist = numpy.array(cont_dists)
			samplerParams.disc.dist = numpy.array(disc_dists)
		_, sampler = verifai.server.choose_sampler(space, samplerType,
		                                           sampler_params=samplerParams)
		self.sampler = sampler

		# default rejection feedback is positive so cross-entropy sampler won't update;
		# for other active samplers an appropriate value should be set manually
		if self.rejectionFeedback is None:
			self.rejectionFeedback = 1

	def nextSample(self, feedback):
		return self.sampler.nextSample(feedback)

	def valueFor(self, param):
		return self.cachedSample[param.index]

class ExternalParameter(Distribution):
	"""A value determined by external code rather than Scenic's internal sampler."""
	def __init__(self):
		super().__init__()
		self.sampler = None
		import scenic.syntax.veneer as veneer	# TODO improve?
		veneer.registerExternalParameter(self)

	def sampleGiven(self, value):
		"""Specialization of  `Samplable.sampleGiven` for external parameters.

		By default, this method simply looks up the value previously sampled by
		`ExternalSampler.sample`.
		"""
		assert self.sampler is not None
		return self.sampler.valueFor(self)

class VerifaiParameter(ExternalParameter):
	"""An external parameter sampled using one of VerifAI's samplers."""

	def __init__(self, domain):
		super().__init__()
		self.domain = domain

	@staticmethod
	def withPrior(dist, buckets=None):
		"""Creates a `VerifaiParameter` using the given distribution as a prior.

		Since the VerifAI cross-entropy sampler currently only supports piecewise-constant
		distributions, if the prior is not of that form it may be approximated. For most
		built-in distributions, the approximation is exact: for a particular distribution,
		check its `bucket` method.
		"""
		if not dist.isPrimitive:
			raise RuntimeError('VerifaiParameter.withPrior called on '
			                   f'non-primitive distribution {dist}')
		bucketed = dist.bucket(buckets=buckets)
		return VerifaiOptions(bucketed.optWeights if bucketed.optWeights else bucketed.options)

class VerifaiRange(VerifaiParameter):
	"""A :obj:`~scenic.core.distributions.Range` (real interval) sampled by VerifAI."""

	def __init__(self, low, high, buckets=None, weights=None):
		import verifai.features
		super().__init__(verifai.features.Box([low, high]))
		if weights is not None:
			weights = tuple(weights)
			if buckets is not None and len(weights) != buckets:
				raise RuntimeError(f'VerifaiRange created with {len(weights)} weights '
				                   f'but {buckets} buckets')
		elif buckets is not None:
			weights = [1] * buckets
		else:
			self.probs = None
			return
		total = sum(weights)
		self.probs = tuple(wt/total for wt in weights)

	def sampleGiven(self, value):
		value = super().sampleGiven(value)
		assert len(value) == 1
		return value[0]

class VerifaiDiscreteRange(VerifaiParameter):
	"""A :obj:`~scenic.core.distributions.DiscreteRange` (integer interval) sampled by VerifAI."""

	def __init__(self, low, high, weights=None):
		import verifai.features
		super().__init__(verifai.features.DiscreteBox([low, high]))
		if weights is not None:
			if len(weights) != (high - low + 1):
				raise RuntimeError(f'VerifaiDiscreteRange created with {len(weights)} weights '
				                   f'for {high - low + 1} values')
			total = sum(weights)
			self.probs = tuple(wt/total for wt in weights)
		else:
			self.probs = None

	def sampleGiven(self, value):
		value = super().sampleGiven(value)
		assert len(value) == 1
		return value[0]

class VerifaiOptions(Options):
	"""An :obj:`~scenic.core.distributions.Options` (discrete set) sampled by VerifAI."""

	@staticmethod
	def makeSelector(n, weights):
		return VerifaiDiscreteRange(0, n, weights)




import json
import itertools
import math
import re

try:
	import pyproj
except ModuleNotFoundError as e:
	raise RuntimeError('guideways scenarios require pyproj to be installed') from e

import numpy as np

from scenic.core.workspaces import Workspace
from scenic.core.regions import PolygonalRegion, PolylineRegion
from scenic.core.vectors import VectorField
from scenic.core.geometry import polygonUnion, headingOfSegment

def toWebots(point, proj):
	x, z = proj(*point)
	return (-x, z)

def localize(points, proj):
	return tuple(toWebots(pt, proj) for pt in points)

class Bordered:
	def __init__(self, proj, data):
		lefts = data['left_border']
		rights = data['right_border']
		points = lefts + list(reversed(rights))
		self.region = PolygonalRegion(localize(points, proj))

	def show(self, plt, style='r'):
		self.region.show(plt, style=style)

class Guideway(Bordered):
	def __init__(self, proj, data):
		super().__init__(proj, data)
		self.id = data['id']
		self.direction = data['direction']
		self.origin_lane = data['origin_lane']['lane_id']
		medianPoints = localize(data['median'], proj)
		self.median = PolylineRegion(medianPoints)
		self.medianPoints = np.array(medianPoints)
		self.region.orientation = VectorField(f'Road{self.id}Direction', self.directionAt)

	def directionAt(self, point):
		# TODO improve?
		median = self.medianPoints
		pt = np.array(point)
		# find closest point on median
		dists = np.linalg.norm(median - pt, axis=1)
		i, dist = min(enumerate(dists), key=lambda p: p[1])
		# if at either end of median, use that direction
		if i == 0:
			x, y = median[:2]
		elif i == len(median)-1:
			x, y = median[-2:]
		else:
			# otherwise, pick which of the two directions to use based on
			# which side of the angle bisector we are on
			d1, d2 = median[i+1] - median[i], median[i-1] - median[i]
			d1 /= np.linalg.norm(d1)
			d2 /= np.linalg.norm(d2)
			bisector = d1 + d2
			if pt[1] * bisector[0] - pt[0] * bisector[1] < 0:
				x, y = median[i-1], median[i]
			else:
				x, y = median[i], median[i+1]
		return headingOfSegment(x, y)

class Crosswalk(Bordered):
	def show(self, plt, style='b'):
		self.region.show(plt, style=style)

class ConflictZone:
	def __init__(self, inter, data):
		pts = localize(data['polygon']['coordinates'][0], inter.projection)
		self.region = PolygonalRegion(pts)
		# TODO not working... some conflict zones have invalid IDs?
		#self.guideway1 = inter.guidewayWithID[data['guideway1_id']]
		#self.guideway2 = inter.guidewayWithID[data['guideway2_id']]

	def show(self, plt):
		self.region.show(plt, style='k')

def projectionAt(point):
	# taken from Webots OSM importer
	lon, lat = point
	utm_zone = 1 + math.floor((float(lon) + 180) / 6)
	hemisphere = 'south' if lat < 0 else 'north'
	string = "+proj=ortho +%s +zone=%d +lon_0=%f +lat_0=%f +x_0=0 +y_0=0 +ellps=WGS84 +units=m +no_defs" % (hemisphere, utm_zone, lon, lat)
	return pyproj.Proj(string)

class Intersection:
	def __init__(self, data, origin=None):
		if origin is None:
			origin = data['main_gw']['median'][0]
		self.projection = projectionAt(origin)

		self.mainGuideway = Guideway(self.projection, data['main_gw'])
		self.vehicleGuideways = (tuple(Guideway(self.projection, d) for d in data['vehicle_gw'])
								 + (self.mainGuideway,))
		self.bikeGuideways = tuple(Guideway(self.projection, d) for d in data['bicycle_gw'])
		self.railGuideways = tuple(Guideway(self.projection, d) for d in data['rail_gw'])
		guidewayWithID = {}
		for gw in itertools.chain(self.vehicleGuideways, self.bikeGuideways,
								  self.railGuideways):
			guidewayWithID[gw.id] = gw
		self.guidewayWithID = guidewayWithID

		self.crosswalks = tuple(Crosswalk(self.projection, d) for d in data['crosswalks'])
		self.conflictZones = tuple(ConflictZone(self, d) for d in data['conflict_zones'])

	def directionAt(self, point):
		for guideway in self.vehicleGuideways:
			if guideway.region.containsPoint(point):
				return guideway.directionAt(point)
		raise RuntimeError('tried to get road direction at point not in a guideway!')

	def show(self, plt):
		for guideway in self.vehicleGuideways:
			guideway.show(plt)
		self.mainGuideway.show(plt, style='g')
		for guideway in self.bikeGuideways:
			guideway.show(plt, style='c')
		for guideway in self.railGuideways:
			guideway.show(plt, style='m')
		for crosswalk in self.crosswalks:
			crosswalk.show(plt)
		for zone in self.conflictZones:
			zone.show(plt)

class IntersectionWorkspace(Workspace):
	def __init__(self, intersection):
		self.intersection = intersection
		drivablePoly = polygonUnion(gw.region.polygons
									for gw in intersection.vehicleGuideways)
		walkablePoly = polygonUnion(cw.region.polygons
									for cw in intersection.crosswalks)
		workspacePoly = polygonUnion((drivablePoly, walkablePoly))
		self.roadDirection = VectorField('RoadDirection', intersection.directionAt)
		self.drivableRegion = PolygonalRegion(polygon=drivablePoly, orientation=self.roadDirection)
		self.workspaceRegion = PolygonalRegion(polygon=workspacePoly)
		super().__init__(self.workspaceRegion)

	def show(self, plt):
		plt.gca().set_aspect('equal')
		self.intersection.show(plt)

	@property
	def minimumZoomSize(self):
		return 30




"""Translator turning Scenic programs into Scenario objects.

The top-level interface to Scenic is provided by two functions:

* `scenarioFromString` -- compile a string of Scenic code;
* `scenarioFromFile` -- compile a Scenic file.

These output a `Scenario` object, from which scenes can be generated.
See the documentation for `Scenario` for details.

When imported, this module hooks the Python import system so that Scenic
modules can be imported using the ``import`` statement. This is primarily for the
translator's own use, but you could import Scenic modules from Python to
inspect them. Because Scenic uses Python's import system, the latter's rules
for finding modules apply, including the handling of packages.

Scenic is compiled in two main steps: translating the code into Python, and
executing the resulting Python module to generate a Scenario object encoding
the objects, distributions, etc. in the scenario. For details, see the function
`compileStream` below.
"""

import sys
import os
import io
import builtins
import traceback
import time
import inspect
import types
import importlib
import importlib.abc
import importlib.util
import itertools
from collections import namedtuple
from contextlib import contextmanager

import tokenize
from tokenize import NAME, NL, NEWLINE, ENDMARKER, OP, NUMBER, COLON, COMMENT, ENCODING
from tokenize import LPAR, RPAR, LSQB, RSQB, COMMA, DOUBLESLASH, DOUBLESLASHEQUAL
from tokenize import AT, LEFTSHIFT, RIGHTSHIFT, VBAR, AMPER, TILDE, CIRCUMFLEX, STAR
from tokenize import LEFTSHIFTEQUAL, RIGHTSHIFTEQUAL, VBAREQUAL, AMPEREQUAL, CIRCUMFLEXEQUAL
from tokenize import INDENT, DEDENT, STRING

import ast
from ast import parse, dump, NodeVisitor, NodeTransformer, copy_location, fix_missing_locations
from ast import Load, Store, Name, Call, Tuple, BinOp, MatMult, BitAnd, BitOr, BitXor, LShift
from ast import RShift, Starred, Lambda, AnnAssign, Set, Str, Num, Subscript, Index

from scenic.core.distributions import Samplable, needsSampling
from scenic.core.lazy_eval import needsLazyEvaluation
from scenic.core.workspaces import Workspace
from scenic.core.scenarios import Scenario
from scenic.core.object_types import Constructible
from scenic.core.utils import ParseError, RuntimeParseError, InvalidScenarioError
import scenic.core.pruning as pruning
import scenic.syntax.veneer as veneer
import scenic.syntax.relations as relations

### THE TOP LEVEL: compiling a Scenic program

def scenarioFromString(string, filename='<string>', cacheImports=False):
	"""Compile a string of Scenic code into a `Scenario`.

	The optional **filename** is used for error messages."""
	stream = io.BytesIO(string.encode())
	return scenarioFromStream(stream, filename=filename, cacheImports=cacheImports)

def scenarioFromFile(path, cacheImports=False):
	"""Compile a Scenic file into a `Scenario`.

	Args:
		path (str): path to a Scenic file
		cacheImports (bool): Whether to cache any imported Scenic modules.
		  The default behavior is to not do this, so that subsequent attempts
		  to import such modules will cause them to be recompiled. If it is
		  safe to cache Scenic modules across multiple compilations, set this
		  argument to True. Then importing a Scenic module will have the same
		  behavior as importing a Python module.

	Returns:
		A `Scenario` object representing the Scenic scenario.
	"""
	if not os.path.exists(path):
		raise FileNotFoundError(path)
	fullpath = os.path.realpath(path)
	head, extension = os.path.splitext(fullpath)
	if not extension or extension[1:] not in scenicExtensions:
		ok = ', '.join(scenicExtensions)
		err = f'Scenic scenario does not have valid extension ({ok})'
		raise RuntimeError(err)
	directory, name = os.path.split(head)

	with open(path, 'rb') as stream:
		return scenarioFromStream(stream, filename=fullpath, path=path,
		                          cacheImports=cacheImports)

def scenarioFromStream(stream, filename='<stream>', path=None, cacheImports=False):
	"""Compile a stream of Scenic code into a `Scenario`."""
	# Compile the code as if it were a top-level module
	oldModules = list(sys.modules.keys())
	try:
		with topLevelNamespace(path) as namespace:
			compileStream(stream, namespace, filename=filename)
	finally:
		if not cacheImports:
			toRemove = []
			for name, module in sys.modules.items():
				if name not in oldModules and getattr(module, '_isScenicModule', False):
					toRemove.append(name)
			for name in toRemove:
				del sys.modules[name]
	# Construct a Scenario from the resulting namespace
	return constructScenarioFrom(namespace)

@contextmanager
def topLevelNamespace(path=None):
	"""Creates an environment like that of a Python script being run directly.

	Specifically, __name__ is '__main__', __file__ is the path used to invoke
	the script (not necessarily its absolute path), and the parent directory is
	added to the path so that 'import blobbo' will import blobbo from that
	directory if it exists there.
	"""
	directory = os.getcwd() if path is None else os.path.dirname(path)
	namespace = { '__name__': '__main__' }
	if path is not None:
		namespace['__file__'] = path
	sys.path.insert(0, directory)
	try:
		yield namespace
	finally:
		del sys.path[0]

def compileStream(stream, namespace, filename='<stream>'):
	"""Compile a stream of Scenic code and execute it in a namespace.

	The compilation procedure consists of the following main steps:

		1. Tokenize the input using the Python tokenizer.
		2. Partition the tokens into blocks separated by import statements.
		   This is done by the `partitionByImports` function.
		3. Translate Scenic constructions into valid Python syntax.
		   This is done by the `TokenTranslator`.
		4. Parse the resulting Python code into an AST using the Python parser.
		5. Modify the AST to achieve the desired semantics for Scenic.
		   This is done by the `translateParseTree` function.
		6. Compile and execute the modified AST.
		7. After executing all blocks, extract the global state (e.g. objects).
		   This is done by the `storeScenarioStateIn` function.
	"""
	if verbosity >= 2:
		veneer.verbosePrint(f'  Compiling Scenic module from {filename}...')
		startTime = time.time()
	# Tokenize input stream
	try:
		tokens = list(tokenize.tokenize(stream.readline))
	except tokenize.TokenError as e:
		line = e.args[1][0] if isinstance(e.args[1], tuple) else e.args[1]
		raise TokenParseError(line, 'file ended during multiline string or expression')
	# Partition into blocks with all imports at the end (since imports could
	# pull in new constructor (Scenic class) definitions, which change the way
	# subsequent tokens are transformed)
	blocks = partitionByImports(tokens)
	veneer.activate()
	newSourceBlocks = []
	try:
		# Execute preamble
		exec(compile(preamble, '<veneer>', 'exec'), namespace)
		# Execute each block
		for blockNum, block in enumerate(blocks):
			# Find all custom constructors defined so far (possibly imported)
			constructors = findConstructorsIn(namespace)
			# Translate tokens to valid Python syntax
			startLine = max(1, block[0][2][0])
			translator = TokenTranslator(constructors)
			newSource, allConstructors = translator.translate(block)
			trimmed = newSource[2*(startLine-1):]	# remove blank lines used to align errors
			newSourceBlocks.append(trimmed)
			if dumpTranslatedPython:
				print(f'### Begin translated Python from block {blockNum} of {filename}')
				print(newSource)
				print('### End translated Python')
			# Parse the translated source
			tree = parseTranslatedSource(newSource, filename)
			# Modify the parse tree to produce the correct semantics
			newTree, requirements = translateParseTree(tree, allConstructors)
			if dumpFinalAST:
				print(f'### Begin final AST from block {blockNum} of {filename}')
				print(ast.dump(newTree, include_attributes=True))
				print('### End final AST')
			# Compile the modified tree
			code = compileTranslatedTree(newTree, filename)
			# Execute it
			executeCodeIn(code, namespace, filename)
		# Extract scenario state from veneer and store it
		storeScenarioStateIn(namespace, requirements, filename)
	finally:
		veneer.deactivate()
	if verbosity >= 2:
		totalTime = time.time() - startTime
		veneer.verbosePrint(f'  Compiled Scenic module in {totalTime:.4g} seconds.')
	allNewSource = ''.join(newSourceBlocks)
	return code, allNewSource

### TRANSLATION PHASE ZERO: definitions of language elements not already in Python

## Options

showInternalBacktrace = False
dumpTranslatedPython = False
dumpFinalAST = False
verbosity = 0
usePruning = True

## Preamble
# (included at the beginning of every module to be translated;
# imports the implementations of the public language features)
preamble = """\
from scenic.syntax.veneer import *
"""

## Get Python names of various elements
## (for checking consistency between the translator and the veneer)

api = set(veneer.__all__)

## Functions used internally

rangeConstructor = 'Range'
createDefault = 'PropertyDefault'
internalFunctions = { rangeConstructor, createDefault }

# sanity check: these functions actually exist
for imp in internalFunctions:
	assert imp in api, imp

## Statements implemented by functions

requireStatement = 'require'
paramStatement = 'param'
functionStatements = { requireStatement, paramStatement, 'mutate' }

# sanity check: implementations of statements actually exist
for imp in functionStatements:
	assert imp in api, imp

## Built-in functions

builtinFunctions = { 'resample', 'verbosePrint' }

# sanity check: implementations of built-in functions actually exist
for imp in builtinFunctions:
	assert imp in api, imp

## Constructors and specifiers

# statement defining a new constructor (Scenic class);
# we still recognize 'constructor' for backwards-compatibility
constructorStatements = ('class', 'constructor')

Constructor = namedtuple('Constructor', ('name', 'parent', 'specifiers'))

pointSpecifiers = {
	('visible', 'from'): 'VisibleFrom',
	('offset', 'by'): 'OffsetBy',
	('offset', 'along'): 'OffsetAlongSpec',
	('at',): 'At',
	('in',): 'In',
	('on',): 'In',
	('beyond',): 'Beyond',
	('visible',): 'VisibleSpec',
	('left', 'of'): 'LeftSpec',
	('right', 'of'): 'RightSpec',
	('ahead', 'of'): 'Ahead',
	('behind',): 'Behind',
	('following',): 'Following',
}
orientedPointSpecifiers = {
	('apparently', 'facing'): 'ApparentlyFacing',
	('facing', 'toward'): 'FacingToward',
	('facing',): 'Facing'
}
objectSpecifiers = {
}

# sanity check: implementations of specifiers actually exist
for imp in pointSpecifiers.values():
	assert imp in api, imp
for imp in orientedPointSpecifiers.values():
	assert imp in api, imp
for imp in objectSpecifiers.values():
	assert imp in api, imp

builtinConstructors = {
	'Point': Constructor('Point', None, pointSpecifiers),
	'OrientedPoint': Constructor('OrientedPoint', 'Point', orientedPointSpecifiers),
	'Object': Constructor('Object', 'OrientedPoint', objectSpecifiers)
}
functionStatements.update(builtinConstructors)

# sanity check: built-in constructors actually exist
for const in builtinConstructors:
	assert const in api, const

## Prefix operators

prefixOperators = {
	('relative', 'position'): 'RelativePosition',
	('relative', 'heading'): 'RelativeHeading',
	('apparent', 'heading'): 'ApparentHeading',
	('distance', 'from'): 'DistanceFrom',
	('distance', 'to'): 'DistanceFrom',
	('angle', 'from'): 'AngleFrom',
	('angle', 'to'): 'AngleTo',
	('ego', '='): 'ego',
	('front', 'left'): 'FrontLeft',
	('front', 'right'): 'FrontRight',
	('back', 'left'): 'BackLeft',
	('back', 'right'): 'BackRight',
	('front',): 'Front',
	('back',): 'Back',
	('left',): 'Left',
	('right',): 'Right',
	('follow',): 'Follow',
	('visible',): 'Visible'
}
assert all(1 <= len(op) <= 2 for op in prefixOperators)
prefixIncipits = { op[0] for op in prefixOperators }
assert not any(op in functionStatements for op in prefixIncipits)

# sanity check: implementations of prefix operators actually exist
for imp in prefixOperators.values():
	assert imp in api, imp

## Infix operators

# pseudo-operator for encoding argument packages for (3+)-ary ops
packageToken = (RIGHTSHIFT, '>>')
packageNode = RShift

InfixOp = namedtuple('InfixOp', ('syntax', 'implementation', 'arity', 'token', 'node'))
infixOperators = (
	# existing Python operators with new semantics
	InfixOp('@', 'Vector', 2, None, MatMult),

	# operators not in Python (in decreasing precedence order)
	InfixOp('at', 'FieldAt', 2, (LEFTSHIFT, '<<'), LShift),
	InfixOp('relative to', 'RelativeTo', 2, (AMPER, '&'), BitAnd),
	InfixOp('offset by', 'RelativeTo', 2, (AMPER, '&'), BitAnd),
	InfixOp('offset along', 'OffsetAlong', 3, (CIRCUMFLEX, '^'), BitXor),
	InfixOp('can see', 'CanSee', 2, (VBAR, '|'), BitOr),

	# just syntactic conveniences, not really operators
	InfixOp('from', None, 2, (COMMA, ','), None),
	InfixOp('for', None, 2, (COMMA, ','), None),
	InfixOp('to', None, 2, (COMMA, ','), None),
	InfixOp('by', None, 2, packageToken, None)
)

infixTokens = {}
infixImplementations = {}
infixIncipits = set()
for op in infixOperators:
	# if necessary, set up map from Scenic to Python syntax
	if op.token is not None:
		tokens = tuple(op.syntax.split(' '))
		assert 1 <= len(tokens) <= 2, op
		assert tokens not in infixTokens, op
		infixTokens[tokens] = op.token
		incipit = tokens[0]
		assert incipit not in functionStatements, op
		infixIncipits.add(incipit)
	# if necessary, set up map from Python to Scenic semantics
	imp = op.implementation
	if imp is not None:
		assert imp in api, op
		node = op.node
		if node in infixImplementations:	# two operators may have the same implementation
			oldArity, oldName = infixImplementations[node]
			assert op.arity == oldArity, (op, oldName)
			assert imp == oldName, (op, oldName)
		else:
			infixImplementations[node] = (op.arity, imp)

allIncipits = prefixIncipits | infixIncipits

## Direct syntax replacements

replacements = {	# TODO police the usage of these? could yield bizarre error messages
	'of': tuple(),
	'deg': ((STAR, '*'), (NUMBER, '0.01745329252')),
	'ego': ((NAME, 'ego'), (LPAR, '('), (RPAR, ')'))
}

## Illegal and reserved syntax

illegalTokens = {
	LEFTSHIFT, RIGHTSHIFT, VBAR, AMPER, TILDE, CIRCUMFLEX,
	LEFTSHIFTEQUAL, RIGHTSHIFTEQUAL, VBAREQUAL, AMPEREQUAL, CIRCUMFLEXEQUAL,
	DOUBLESLASH, DOUBLESLASHEQUAL
}

# sanity check: stand-in tokens for infix operators must be illegal
for token in infixTokens.values():
	ttype = token[0]
	assert (ttype is COMMA or ttype in illegalTokens), token

keywords = (set(constructorStatements)
	| internalFunctions | functionStatements
	| replacements.keys())

### TRANSLATION PHASE ONE: handling imports

## Meta path finder and loader for Scenic files

scenicExtensions = ('sc', 'scenic')

class ScenicMetaFinder(importlib.abc.MetaPathFinder):
	def find_spec(self, name, paths, target):
		if paths is None:
			paths = sys.path
			modname = name
		else:
			modname = name.rpartition('.')[2]
		for path in paths:
			for extension in scenicExtensions:
				filename = modname + '.' + extension
				filepath = os.path.join(path, filename)
				if os.path.exists(filepath):
					filepath = os.path.abspath(filepath)
					spec = importlib.util.spec_from_file_location(name, filepath,
						loader=ScenicLoader(filepath, filename))
					return spec
		return None

class ScenicLoader(importlib.abc.InspectLoader):
	def __init__(self, filepath, filename):
		self.filepath = filepath
		self.filename = filename

	def create_module(self, spec):
		return None

	def exec_module(self, module):
		# Read source file and compile it
		with open(self.filepath, 'r') as stream:
			source = stream.read()
		with open(self.filepath, 'rb') as stream:
			code, pythonSource = compileStream(stream, module.__dict__, filename=self.filepath)
		# Mark as a Scenic module
		module._isScenicModule = True
		# Save code, source, and translated source for later inspection
		module._code = code
		module._source = source
		module._pythonSource = pythonSource

	def is_package(self, fullname):
		return False

	def get_code(self, fullname):
		module = importlib.import_module(fullname)
		assert module._isScenicModule, module
		return module._code

	def get_source(self, fullname):
		module = importlib.import_module(fullname)
		assert module._isScenicModule, module
		return module._pythonSource

# register the meta path finder
sys.meta_path.insert(0, ScenicMetaFinder())

## Post-import hook to inherit objects, etc. from imported Scenic modules

def hooked_import(*args, **kwargs):
	"""Version of __import__ hooked by Scenic to capture Scenic modules."""
	module = original_import(*args, **kwargs)
	if getattr(module, '_isScenicModule', False):
		if veneer.isActive():
			veneer.allObjects.extend(module._objects)
			veneer.globalParameters.update(module._params)
			veneer.externalParameters.extend(module._externalParams)
			veneer.inheritedReqs.extend(module._requirements)
	return module

original_import = builtins.__import__
builtins.__import__ = hooked_import

## Miscellaneous utilities

def partitionByImports(tokens):
	"""Partition the tokens into blocks ending with import statements."""
	blocks = []
	currentBlock = []
	duringImport = False
	haveImported = False
	finishLine = False
	parenLevel = 0
	for token in tokens:
		startNewBlock = False
		if token.exact_type == LPAR:
			parenLevel += 1
		elif token.exact_type == RPAR:
			parenLevel -= 1
		if finishLine:
			if token.type in (NEWLINE, NL) and parenLevel == 0:
				finishLine = False
				if duringImport:
					duringImport = False
					haveImported = True
		else:
			assert not duringImport
			finishLine = True
			if token.type == NAME and token.string == 'import' or token.string == 'from':
				duringImport = True
			elif token.type in (NEWLINE, NL, COMMENT, ENCODING):
				finishLine = False
			elif haveImported:
				# could use new constructors; needs to be in a new block
				startNewBlock = True
		if startNewBlock:
			blocks.append(currentBlock)
			currentBlock = [token]
			haveImported = False
		else:
			currentBlock.append(token)
	blocks.append(currentBlock)		# add last block
	return blocks

def findConstructorsIn(namespace):
	"""Find all constructors (Scenic classes) defined in a namespace."""
	constructors = []
	for name, value in namespace.items():
		if inspect.isclass(value) and issubclass(value, Constructible):
			if name in builtinConstructors:
				continue
			parent = None
			for base in value.__bases__:
				if issubclass(base, Constructible):
					assert parent is None
					parent = base
			constructors.append(Constructor(name, parent.__name__, {}))
	return constructors

### TRANSLATION PHASE TWO: translation at the level of tokens

class TokenParseError(ParseError):
	"""Parse error occurring during token translation."""
	def __init__(self, tokenOrLine, message):
		line = tokenOrLine.start[0] if hasattr(tokenOrLine, 'start') else tokenOrLine
		self.lineno = line
		super().__init__('Parse error in line ' + str(line) + ': ' + message)

class Peekable:
	"""Utility class to allow iterator lookahead."""
	def __init__(self, gen):
		self.gen = iter(gen)
		self.current = next(self.gen, None)

	def __iter__(self):
		return self

	def __next__(self):
		cur = self.current
		if cur is None:
			raise StopIteration
		self.current = next(self.gen, None)
		return cur

	def peek(self):
		return self.current

def peek(thing):
	return thing.peek()

class TokenTranslator:
	"""Translates a Scenic token stream into valid Python syntax.

	This is a stateful process because constructor (Scenic class) definitions
	change the way subsequent code is parsed.
	"""
	def __init__(self, constructors=()):
		self.functions = set(functionStatements)
		self.constructors = dict(builtinConstructors)
		for constructor in constructors:
			name = constructor.name
			assert name not in self.constructors
			self.constructors[name] = constructor
			self.functions.add(name)

	def createConstructor(self, name, parent, specs={}):
		if parent is None:
			parent = 'Object'		# default superclass
		self.constructors[name] = Constructor(name, parent, specs)
		self.functions.add(name)
		return parent

	def specifiersForConstructor(self, const):
		name, parent, specs = self.constructors[const]
		if parent is None:
			return specs
		else:
			ps = dict(self.specifiersForConstructor(parent))
			ps.update(specs)
			return ps

	def translate(self, tokens):
		"""Do the actual translation of the token stream."""
		tokens = Peekable(tokens)
		newTokens = []
		functionStack = []
		inConstructor = False	# inside a constructor or one of its specifiers
		specifiersIndented = False
		parenLevel = 0
		row, col = 0, 0		# position of next token to write out
		orow, ocol = 0, 0	# end of last token in the original source
		startOfLine = True		# TODO improve hack?
		functions = self.functions
		constructors = self.constructors

		for token in tokens:
			ttype = token.exact_type
			tstring = token.string
			skip = False
			endToken = token 	# token to advance past in column count
			movedUpTo = False

			def injectToken(tok, spaceAfter=0):
				"""Add a token to the output stream, trying to preserve spacing."""
				nonlocal row, col, movedUpTo
				if not movedUpTo:
					moveUpTo(token)
					moveBeyond(token)
					movedUpTo = True
				ty, string = tok[:2]
				if len(tok) >= 3:
					moveBeyond(tok)
					srow, scol = tok[2]
					erow, ecol = tok[3]
					width = ecol - scol
					height = erow - srow
				else:
					width = len(string)
					height = 0
				ncol = ecol if height > 0 else col + width
				newToken = (ty, string, (row, col), (row+height, ncol), '')
				newTokens.append(newToken)
				if ty in (NEWLINE, NL):
					row += 1
					col = 0
				elif height > 0:
					row += height
					col = ncol
				else:
					col += width + spaceAfter
			def moveUpTo(tok):
				nonlocal row, col, orow, ocol
				nrow, ncol = tok[2]
				if nrow > orow:
					row = nrow
					col = ncol
				else:
					gap = ncol - ocol
					assert gap >= 0, (tok, row, col, ocol)
					col += gap
			def moveBeyond(tok):
				nonlocal orow, ocol
				nrow, ncol = tok[3]
				if nrow > orow or (nrow == orow and ncol > ocol):
					orow = nrow
					ocol = ncol
			def advance(skip=True):
				nextToken = next(tokens)
				if skip:
					moveBeyond(nextToken)
				else:
					injectToken(nextToken)
				return nextToken
			def callFunction(function, argument=None):
				nonlocal skip, matched, functionStack
				functionStack.append((function, parenLevel))
				injectToken((NAME, function))
				injectToken((LPAR, '('))
				if argument is not None:
					injectToken((NAME, argument))
					injectToken((COMMA, ','))
				skip = True
				matched = True

			# Catch Python operators that can't be used in Scenic
			if ttype in illegalTokens:
				raise TokenParseError(token, f'illegal operator "{tstring}"')

			# Determine which operators are allowed in current context
			context, startLevel = functionStack[-1] if functionStack else (None, None)
			inConstructorContext = (context in constructors and parenLevel == startLevel)
			if inConstructorContext:
				inConstructor = True
				allowedPrefixOps = self.specifiersForConstructor(context)
				allowedInfixOps = dict()
			else:
				allowedPrefixOps = prefixOperators
				allowedInfixOps = infixTokens

			# Parse next token
			if ttype == LPAR or ttype == LSQB:		# keep track of nesting level
				parenLevel += 1
			elif ttype == RPAR or ttype == RSQB:	# ditto
				parenLevel -= 1
			elif ttype == STRING:
				# special case for global parameters with quoted names:
				# transform "name"=value into "name", value
				if (len(functionStack) > 0 and functionStack[-1][0] == paramStatement
				    and peek(tokens).string == '='):
					next(tokens)	# consume '='
					injectToken(token)
					injectToken((COMMA, ','))
					skip = True
			elif ttype == NAME:		# the interesting case: almost all new syntax falls in here
				# try to match 2-word language constructs
				matched = False
				nextToken = peek(tokens)		# lookahead so we can give 2-word ops precedence
				if nextToken is not None:
					endToken = nextToken	# tentatively; will be overridden if no match
					nextString = nextToken.string
					twoWords = (tstring, nextString)
					if startOfLine and tstring == 'for':	# TODO improve hack?
						matched = True
						endToken = token
					elif startOfLine and tstring in constructorStatements:	# class definition
						if nextToken.type != NAME or nextString in keywords:
							raise TokenParseError(nextToken,
							    f'invalid class name "{nextString}"')
						nextToken = next(tokens)	# consume name
						parent = None
						pythonClass = False
						if peek(tokens).exact_type == LPAR:		# superclass specification
							next(tokens)
							nextToken = next(tokens)
							parent = nextToken.string
							if nextToken.exact_type != NAME or parent in keywords:
								raise TokenParseError(nextToken,
								    f'invalid superclass "{parent}"')
							if parent not in self.constructors:
								if tstring != 'class':
									raise TokenParseError(nextToken,
									    f'superclass "{parent}" is not a Scenic class')
								# appears to be a Python class definition
								pythonClass = True
							else:
								nextToken = next(tokens)
								if nextToken.exact_type != RPAR:
									raise TokenParseError(nextToken,
									                      'malformed class definition')
						injectToken((NAME, 'class'), spaceAfter=1)
						injectToken((NAME, nextString))
						injectToken((LPAR, '('))
						if pythonClass:		# pass Python class definitions through unchanged
							while nextToken.exact_type != COLON:
								injectToken(nextToken)
								nextToken = next(tokens)
							injectToken(nextToken)
						else:
							if peek(tokens).exact_type != COLON:
								raise TokenParseError(nextToken, 'malformed class definition')
							parent = self.createConstructor(nextString, parent)
							injectToken((NAME, parent))
							injectToken((RPAR, ')'))
						skip = True
						matched = True
						endToken = nextToken
					elif twoWords in allowedPrefixOps:	# 2-word prefix operator
						callFunction(allowedPrefixOps[twoWords])
						advance()	# consume second word
					elif not startOfLine and twoWords in allowedInfixOps:	# 2-word infix operator
						injectToken(allowedInfixOps[twoWords])
						advance()	# consume second word
						skip = True
						matched = True
					elif inConstructorContext and tstring == 'with':	# special case for 'with' specifier
						callFunction('With', argument=f'"{nextString}"')
						advance()	# consume property name
					elif tstring == requireStatement and nextString == '[':		# special case for require[p]
						next(tokens)	# consume '['
						nextToken = next(tokens)
						if nextToken.exact_type != NUMBER:
							raise TokenParseError(nextToken,
							    'soft requirement must have constant probability')
						prob = nextToken.string
						nextToken = next(tokens)
						if nextToken.exact_type != RSQB:
							raise TokenParseError(nextToken, 'malformed soft requirement')
						callFunction(requireStatement, argument=prob)
						endToken = nextToken
				if not matched:
					# 2-word constructs don't match; try 1-word
					endToken = token
					oneWord = (tstring,)
					if oneWord in allowedPrefixOps:		# 1-word prefix operator
						callFunction(allowedPrefixOps[oneWord])
					elif not startOfLine and oneWord in allowedInfixOps:	# 1-word infix operator
						injectToken(allowedInfixOps[oneWord])
						skip = True
					elif inConstructorContext:		# couldn't match any 1- or 2-word specifier
						raise TokenParseError(token, f'unknown specifier "{tstring}"')
					elif tstring in functions:		# built-in function
						callFunction(tstring)
					elif tstring in replacements:	# direct replacement
						for tok in replacements[tstring]:
							injectToken(tok)
						skip = True
					elif startOfLine and tstring == 'from':		# special case to allow 'from X import Y'
						pass
					elif tstring in keywords:		# some malformed usage
						raise TokenParseError(token, f'unexpected keyword "{tstring}"')
					else:
						pass	# nothing matched; pass through unchanged to Python

			# Detect the end of function argument lists
			if len(functionStack) > 0:
				context, startLevel = functionStack[-1]
				while parenLevel < startLevel:		# we've closed all parens for the current function
					functionStack.pop()
					injectToken((RPAR, ')'))
					context, startLevel = (None, 0) if len(functionStack) == 0 else functionStack[-1]
				if inConstructor and parenLevel == startLevel and ttype == COMMA:		# starting a new specifier
					while functionStack and context not in constructors:
						functionStack.pop()
						injectToken((RPAR, ')'))
						context, startLevel = (None, 0) if len(functionStack) == 0 else functionStack[-1]
					# allow the next specifier to be on the next line, if indented
					injectToken(token)		# emit comma immediately
					skip = True
					nextToken = peek(tokens)
					specOnNewLine = False
					while nextToken.exact_type in (NEWLINE, NL, COMMENT, ENDMARKER):
						specOnNewLine = True
						if nextToken.exact_type == COMMENT:
							advance(skip=False)		# preserve comment
							nextToken = peek(tokens)
						if nextToken.exact_type not in (NEWLINE, NL):
							raise TokenParseError(nextToken, 'comma with no specifier following')
						advance(skip=False)		# preserve newline
						nextToken = peek(tokens)
					if specOnNewLine and not specifiersIndented:
						nextToken = next(tokens)		# consume indent
						if nextToken.exact_type != INDENT:
							raise TokenParseError(nextToken,
							                      'expected indented specifier (extra comma on previous line?)')
						injectToken(nextToken)
						specifiersIndented = True
				elif ttype == NEWLINE or ttype == ENDMARKER or ttype == COMMENT:	# end of line
					inConstructor = False
					if parenLevel != 0:
						raise TokenParseError(token, 'unmatched parens/brackets')
					while len(functionStack) > 0:
						functionStack.pop()
						injectToken((RPAR, ')'))

			# Output token unchanged, unless handled above
			if not skip:
				injectToken(token)
			else:
				moveBeyond(endToken)
			startOfLine = (ttype in (ENCODING, NEWLINE, NL, INDENT, DEDENT))

		rewrittenSource = tokenize.untokenize(newTokens)
		if not isinstance(rewrittenSource, str):	# TODO improve?
			rewrittenSource = str(rewrittenSource, encoding='utf-8')
		return rewrittenSource, self.constructors

### TRANSLATION PHASE THREE: parsing of Python resulting from token translation

class PythonParseError(SyntaxError, ParseError):
	"""Parse error occurring during Python parsing or compilation."""
	@classmethod
	def fromSyntaxError(cls, exc):
		msg, (filename, lineno, offset, line) = exc.args
		try:	# attempt to recover line from original file
			with open(filename, 'r') as f:
				line = list(itertools.islice(f, lineno-1, lineno))
			assert len(line) == 1
			line = line[0]
			offset = min(offset, len(line))		# TODO improve?
		except FileNotFoundError:
			pass
		newExc = cls(msg, (filename, lineno, offset, line))
		return newExc.with_traceback(exc.__traceback__)

def parseTranslatedSource(source, filename):
	try:
		tree = parse(source, filename=filename)
		return tree
	except SyntaxError as e:
		cause = e if showInternalBacktrace else None
		raise PythonParseError.fromSyntaxError(e) from cause

### TRANSLATION PHASE FOUR: modifying the parse tree

noArgs = ast.arguments(
	args=[], vararg=None,
	kwonlyargs=[], kw_defaults=[],
	kwarg=None, defaults=[])
selfArg = ast.arguments(
	args=[ast.arg(arg='self', annotation=None)], vararg=None,
	kwonlyargs=[], kw_defaults=[],
	kwarg=None, defaults=[])

if sys.version_info >= (3, 8):	# TODO cleaner way to handle this?
	noArgs.posonlyargs = []
	selfArg.posonlyargs = []

class AttributeFinder(NodeVisitor):
	"""Utility class for finding all referenced attributes of a given name."""
	@staticmethod
	def find(target, node):
		af = AttributeFinder(target)
		af.visit(node)
		return af.attributes

	def __init__(self, target):
		super().__init__()
		self.target = target
		self.attributes = set()

	def visit_Attribute(self, node):
		val = node.value
		if isinstance(val, Name) and val.id == self.target:
			self.attributes.add(node.attr)
		self.visit(val)

class ASTParseError(ParseError):
	"""Parse error occuring during modification of the Python AST."""
	def __init__(self, line, message):
		self.lineno = line
		super().__init__('Parse error in line ' + str(line) + ': ' + message)

class ASTSurgeon(NodeTransformer):
	def __init__(self, constructors):
		super().__init__()
		self.constructors = set(constructors.keys())
		self.requirements = []

	def parseError(self, node, message):
		raise ASTParseError(node.lineno, message)

	def unpack(self, arg, expected, node):
		"""Unpack arguments to ternary (and up) infix operators."""
		assert expected > 0
		if isinstance(arg, BinOp) and isinstance(arg.op, packageNode):
			if expected == 1:
				raise self.parseError(node, 'gave too many arguments to infix operator')
			else:
				return self.unpack(arg.left, expected - 1, node) + [self.visit(arg.right)]
		elif expected > 1:
			raise self.parseError(node, 'gave too few arguments to infix operator')
		else:
			return [self.visit(arg)]

	def visit_BinOp(self, node):
		"""Convert infix operators to calls to the corresponding Scenic operator implementations."""
		left = node.left
		right = node.right
		op = node.op
		if isinstance(op, packageNode):		# unexpected argument package
			raise self.parseError(node, 'unexpected keyword "by"')
		elif type(op) in infixImplementations:	# an operator with non-Python semantics
			arity, impName = infixImplementations[type(op)]
			implementation = Name(impName, Load())
			copy_location(implementation, node)
			assert arity >= 2
			args = [self.visit(left)] + self.unpack(right, arity-1, node)
			newNode = Call(implementation, args, [])
		else:	# all other operators have the Python semantics
			newNode = BinOp(self.visit(left), op, self.visit(right))
		return copy_location(newNode, node)

	def visit_Tuple(self, node):
		"""Convert pairs into uniform distributions."""
		if isinstance(node.ctx, Store):
			return self.generic_visit(node)
		if len(node.elts) != 2:
			raise self.parseError(node, 'interval must have exactly two endpoints')
		newElts = [self.visit(elt) for elt in node.elts]
		return copy_location(Call(Name(rangeConstructor, Load()), newElts, []), node)

	def visit_Call(self, node):
		"""Wrap require statements with lambdas and unpack any argument packages."""
		func = node.func
		if isinstance(func, Name) and func.id == requireStatement:	# Require statement
			# Soft reqs have 2 arguments, including the probability, which is given as the
			# first argument by the token translator; so we allow an extra argument here and
			# validate it later on (in case the user wrongly gives 2 arguments to require).
			if not (1 <= len(node.args) <= 2):
				raise self.parseError(node, 'require takes exactly one argument')
			if len(node.keywords) != 0:
				raise self.parseError(node, 'require takes no keyword arguments')
			cond = node.args[-1]
			if isinstance(cond, Starred):
				raise self.parseError(node, 'argument unpacking cannot be used with require')
			req = self.visit(cond)
			reqID = Num(len(self.requirements))	# save ID number
			self.requirements.append(req)		# save condition for later inspection when pruning
			closure = Lambda(noArgs, req)		# enclose requirement in a lambda
			lineNum = Num(node.lineno)			# save line number for error messages
			copy_location(closure, req)
			copy_location(lineNum, req)
			newArgs = [reqID, closure, lineNum]
			if len(node.args) == 2:		# get probability for soft requirements
				prob = node.args[0]
				if not isinstance(prob, Num):
					raise self.parseError(node, 'malformed requirement '
					                            '(should be a single expression)')
				newArgs.append(prob)
			return copy_location(Call(func, newArgs, []), node)
		else:	# Ordinary function call
			newFunc = self.visit(func)
			newArgs = []
			# Translate arguments, unpacking any argument packages
			for arg in node.args:
				if isinstance(arg, BinOp) and isinstance(arg.op, packageNode):
					newArgs.extend(self.unpack(arg, 2, node))
				else:
					newArgs.append(self.visit(arg))
			newKeywords = [self.visit(kwarg) for kwarg in node.keywords]
			return copy_location(Call(newFunc, newArgs, newKeywords), node)

	def visit_ClassDef(self, node):
		"""Process property defaults for Scenic classes."""
		if node.name in self.constructors:		# Scenic class definition
			newBody = []
			for child in node.body:
				child = self.visit(child)
				if isinstance(child, AnnAssign):	# default value for property
					origValue = child.annotation
					target = child.target
					# extract any attributes for this property
					metaAttrs = []
					if isinstance(target, Subscript):
						sl = target.slice
						if not isinstance(sl, Index):
							self.parseError(sl, 'malformed attributes for property default')
						sl = sl.value
						if isinstance(sl, Name):
							metaAttrs.append(sl.id)
						elif isinstance(sl, Tuple):
							for elt in sl.elts:
								if not isinstance(elt, Name):
									self.parseError(elt,
									    'malformed attributes for property default')
								metaAttrs.append(elt.id)
						else:
							self.parseError(sl, 'malformed attributes for property default')
						newTarget = Name(target.value.id, Store())
						copy_location(newTarget, target)
						target = newTarget
					# find dependencies of the default value
					properties = AttributeFinder.find('self', origValue)
					# create default value object
					args = [
						Set([Str(prop) for prop in properties]),
						Set([Str(attr) for attr in metaAttrs]),
						Lambda(selfArg, origValue)
					]
					value = Call(Name(createDefault, Load()), args, [])
					copy_location(value, origValue)
					newChild = AnnAssign(
						target=target, annotation=value,
						value=None, simple=True)
					child = copy_location(newChild, child)
				newBody.append(child)
			node.body = newBody
			return node
		else:		# ordinary Python class
			# it's impossible at the moment to define a Python class in a Scenic file,
			# but we'll leave this check here for future-proofing
			for base in node.bases:
				name = None
				if isinstance(base, Call):
					name = base.func.id
				elif isinstance(base, Name):
					name = base.id
				if name is not None and name in self.constructors:
					self.parseError(node,
					                f'Python class {node.name} derives from Scenic class {name}')
			return self.generic_visit(node)

def translateParseTree(tree, constructors):
	"""Modify the Python AST to produce the desired Scenic semantics."""
	surgeon = ASTSurgeon(constructors)
	tree = fix_missing_locations(surgeon.visit(tree))
	return tree, surgeon.requirements

### TRANSLATION PHASE FIVE: AST compilation

def compileTranslatedTree(tree, filename):
	try:
		return compile(tree, filename, 'exec')
	except SyntaxError as e:
		cause = e if showInternalBacktrace else None
		raise PythonParseError.fromSyntaxError(e) from cause

### TRANSLATION PHASE SIX: Python execution

def generateTracebackFrom(exc, sourceFile):
	"""Trim an exception's traceback to the last line of Scenic code."""
	# find last stack frame in the source file
	tbexc = traceback.TracebackException.from_exception(exc)
	last = None
	tbs = []
	currentTb = exc.__traceback__
	for depth, frame in enumerate(tbexc.stack):
		assert currentTb is not None
		tbs.append(currentTb)
		currentTb = currentTb.tb_next
		if frame.filename == sourceFile:
			last = depth
	assert last is not None

	# create new trimmed traceback
	lastTb = tbs[last]
	lastLine = lastTb.tb_lineno
	tbs = tbs[:last]
	try:
		currentTb = types.TracebackType(None, lastTb.tb_frame,
		                                lastTb.tb_lasti, lastLine)
	except TypeError:
		# Python 3.6 does not allow creation of traceback objects, so we just
		# return the original traceback
		return exc.__traceback__, lastLine

	for tb in reversed(tbs):
		currentTb = types.TracebackType(currentTb, tb.tb_frame,
		                                tb.tb_lasti, tb.tb_lineno)
	return currentTb, lastLine

class InterpreterParseError(ParseError):
	"""Parse error occuring during Python execution."""
	def __init__(self, exc, line):
		self.lineno = line
		exc_name = type(exc).__name__
		super().__init__(f'Parse error in line {line}: {exc_name}: {exc}')

def executeCodeIn(code, namespace, filename):
	"""Execute the final translated Python code in the given namespace."""
	executePythonFunction(lambda: exec(code, namespace), filename)

def executePythonFunction(func, filename):
	"""Execute a Python function, giving correct Scenic backtraces for any exceptions."""
	try:
		return func()
	except RuntimeParseError as e:
		cause = e if showInternalBacktrace else None
		tb, line = generateTracebackFrom(e, filename)
		raise InterpreterParseError(e, line).with_traceback(tb) from cause

### TRANSLATION PHASE SEVEN: scenario construction

def storeScenarioStateIn(namespace, requirementSyntax, filename):
	"""Post-process an executed Scenic module, extracting state from the veneer."""
	# Extract created Objects
	namespace['_objects'] = tuple(veneer.allObjects)
	namespace['_egoObject'] = veneer.egoObject

	# Extract global parameters
	namespace['_params'] = veneer.globalParameters
	for name, value in veneer.globalParameters.items():
		if needsLazyEvaluation(value):
			raise InvalidScenarioError(f'parameter {name} uses value {value}'
			                           ' undefined outside of object definition')

	# Extract external parameters
	namespace['_externalParams'] = tuple(veneer.externalParameters)

	# Extract requirements, scan for relations used for pruning, and create closures
	requirements = veneer.pendingRequirements
	finalReqs = veneer.inheritedReqs
	requirementDeps = set()	# things needing to be sampled to evaluate the requirements
	namespace['_requirements'] = finalReqs
	namespace['_requirementDeps'] = requirementDeps
	def makeClosure(req, bindings, ego, line):
		"""Create a closure testing the requirement in the correct runtime state."""
		def evaluator():
			result = req()
			assert not needsSampling(result)
			if needsLazyEvaluation(result):
				raise InvalidScenarioError(f'requirement on line {line} uses value'
				                           ' undefined outside of object definition')
			return result
		def closure(values):
			# rebind any names referring to sampled objects
			for name, value in bindings.items():
				if value in values:
					namespace[name] = values[value]
			# evaluate requirement condition, reporting errors on the correct line
			try:
				veneer.evaluatingRequirement = True
				# rebind ego object, which can be referred to implicitly
				assert veneer.egoObject is None
				if ego is not None:
					veneer.egoObject = values[ego]
				result = executePythonFunction(evaluator, filename)
			finally:
				veneer.evaluatingRequirement = False
				veneer.egoObject = None
			return result
		return closure
	for reqID, (req, bindings, ego, line, prob) in requirements.items():
		# Check whether requirement implies any relations used for pruning
		reqNode = requirementSyntax[reqID]
		relations.inferRelationsFrom(reqNode, bindings, ego, line)
		# Gather dependencies of the requirement
		for value in bindings.values():
			if needsSampling(value):
				requirementDeps.add(value)
			if needsLazyEvaluation(value):
				raise InvalidScenarioError(f'requirement on line {line} uses value {value}'
				                           ' undefined outside of object definition')
		if ego is not None:
			assert isinstance(ego, Samplable)
			requirementDeps.add(ego)
		# Construct closure
		finalReqs.append((makeClosure(req, bindings, ego, line), prob))

def constructScenarioFrom(namespace):
	"""Build a Scenario object from an executed Scenic module."""
	# Extract ego object
	if namespace['_egoObject'] is None:
		raise InvalidScenarioError('did not specify ego object')

	# Extract workspace, if one is specified
	if 'workspace' in namespace:
		workspace = namespace['workspace']
		if not isinstance(workspace, Workspace):
			raise InvalidScenarioError(f'workspace {workspace} is not a Workspace')
		if needsSampling(workspace):
			raise InvalidScenarioError('workspace must be a fixed region')
		if needsLazyEvaluation(workspace):
			raise InvalidScenarioError('workspace uses value undefined '
			                           'outside of object definition')
	else:
		workspace = None

	# Create Scenario object
	scenario = Scenario(workspace,
	                    namespace['_objects'], namespace['_egoObject'],
	                    namespace['_params'], namespace['_externalParams'],
	                    namespace['_requirements'], namespace['_requirementDeps'])

	# Prune infeasible parts of the space
	if usePruning:
		pruning.prune(scenario, verbosity=verbosity)

	return scenario



"""Stub to allow changing the Webots world without changing the model."""

import os.path

#: Path to the WBT file to load the Webots world from
worldPath = None

def setLocalWorld(module, relpath):
    """Select a WBT file relative to the given module.

    This function is intended to be used with ``__file__`` as the *module*.
    """
    global worldPath
    base = os.path.dirname(module)
    worldPath = os.path.join(base, relpath)



'''
This file has basic image modification functions
'''

from PIL import Image
import cv2
from scipy.spatial import Voronoi
from itertools import product
import numpy as np


def convert_black_white(img_data=None, img_file=None, threshold=100):
    assert img_data is not None or img_file is not None
    if img_data is None:
        img_data = Image.open(img_file)

    img_copy = img_data.copy()
    pixels = img_copy.load()

    for j,k in product(range(img_copy.size[0]), range(img_copy.size[1])):
        if (np.array(pixels[j, k][0:3]) > threshold).any():
            pixels[j, k] = (255, 255, 255, 255)
        else:
            pixels[j,k] = (0, 0, 0, 255)

    return img_copy

def get_edges(img_data=None, img_file=None, threshold=100, kernelsize=1):
    assert img_data is not None or img_file is not None
    if img_data is None:
        img_data = Image.open(img_file)

    img_copy = img_data.copy()

    # Get the black and white image
    img_bw = convert_black_white(img_data=img_copy, img_file=img_file,
                                 threshold=threshold)
    cv_bw = cv2.cvtColor(np.array(img_bw), cv2.COLOR_RGB2BGR)
    # Detect edges using Laplacian
    laplacian = cv2.Laplacian(cv_bw, cv2.CV_8U, ksize=kernelsize)

    # Convert back to Pillow image
    pil_lap = Image.fromarray(laplacian)


    # For computing Voronoi images, we need to squeeze the RGB data to 0s and 1s
    pil_squeezed = pil_lap.convert('L')
    pil_squeezed_01 = pil_squeezed.point(lambda x: 0 if x < 128 else 255, '1')
    return pil_squeezed_01

def voronoi_edge(img_data=None, img_file=None, threshold=100, kernelsize=1):
    assert img_data is not None or img_file is not None
    if img_data is None:
        img_data = Image.open(img_file)

    img_copy = img_data.copy()

    # Get 0s and 1s of the edges
    pil_squeezed_01 = get_edges(img_data=img_copy, img_file=img_file,
                                threshold=threshold, kernelsize=kernelsize)

    # Collecting point for Voronoi edge computation
    nz_elements = np.nonzero(np.asarray(pil_squeezed_01))
    points = np.fliplr(np.array(nz_elements).T)
    vor = Voronoi(points)
    vor_x = vor.vertices.T[0]
    vor_y = -vor.vertices.T[1] + img_data.size[1]

    # Convert the black and white image to 0s and 1s
    img_bw = convert_black_white(img_data=img_copy,
                                 img_file=img_file, threshold=threshold)
    img_bw_squeezed = img_bw.convert('L')
    img_bw_01 = img_bw_squeezed.point(lambda x:0 if x< 128 else 255, '1')
    pixels = img_bw_01.load()

    center_x = []
    center_y = []
    for x, y in zip(vor_x, vor_y):
        if 0 < x and x < img_data.size[0] and 0 < y and y < img_data.size[1] \
                and pixels[int(x), img_data.size[1]-1 -int(y)] == 0:
            center_x.append(int(x))
            center_y.append(int(y))

    return {'edge_image':pil_squeezed_01, 'vor_center_x': center_x,
            'vor_center_y': center_y}


def plot_voronoi_plot(img_data=None, img_file=None, threshold=100, kernelsize=3,
    plot_name=None):
    import matplotlib.pyplot as plt
    assert img_data is not None or img_file is not None
    vor_results = voronoi_edge(img_data=img_data, img_file=img_file,
                               threshold=threshold, kernelsize=kernelsize)

    xlim = vor_results['edge_image'].size[0]
    ylim = vor_results['edge_image'].size[1]

    x_data = vor_results['vor_center_x']
    y_data = vor_results['vor_center_y']
    plt.figure()
    plt.scatter(x_data, y_data, s=0.5)
    plt.xlim(0, xlim)
    plt.ylim(0, ylim)
    if plot_name is None:
        plt.savefig('voronoi_fig.png')
    else:
        plt.savefig(plot_name+'.png')






"""A compiler and scene generator for the Scenic scenario description language.

.. raw:: html

   <h2>Submodules</h2>

.. autosummary::
   :toctree: _autosummary

   core
   simulators
   syntax
"""

from .syntax.translator import scenarioFromFile, scenarioFromString



"""Scenic world model for Grand Theft Auto V (GTAV).

.. raw:: html

   <h2>Submodules</h2>

.. autosummary::
   :toctree: _autosummary

   model
   interface
   center_detection
   img_modf
   map
   messages
"""



"""Scenic world models for the Webots robotics simulator.

This module contains common code for working with Webots, e.g. parsing WBT files.
World models for particular uses of Webots are in submodules.

.. raw:: html

   <h2>Submodules</h2>

.. autosummary::
   :toctree: _autosummary

   mars
   road
   guideways
   common
   world_parser
"""

from .common import scenicToWebotsPosition, scenicToWebotsRotation



"""Utility functions for geometric computation."""

import math
import itertools
import warnings

import numpy as np
import shapely.geometry
import shapely.ops
import pypoly2tri

from scenic.core.distributions import (needsSampling, distributionFunction,
                                       monotonicDistributionFunction)
from scenic.core.lazy_eval import needsLazyEvaluation
import scenic.core.utils as utils

@distributionFunction
def sin(x):
	return math.sin(x)

@distributionFunction
def cos(x):
	return math.cos(x)

@monotonicDistributionFunction
def hypot(x, y):
	return math.hypot(x, y)

@monotonicDistributionFunction
def max(*args):
	return __builtins__['max'](*args)

@monotonicDistributionFunction
def min(*args):
	return __builtins__['min'](*args)

def normalizeAngle(angle):
	while angle > math.pi:
		angle -= math.tau
	while angle < -math.pi:
		angle += math.tau
	assert -math.pi <= angle <= math.pi
	return angle

def addVectors(a, b):
	ax, ay = a[0], a[1]
	bx, by = b[0], b[1]
	return (ax + bx, ay + by)

def subtractVectors(a, b):
	ax, ay = a[0], a[1]
	bx, by = b[0], b[1]
	return (ax - bx, ay - by)

def averageVectors(a, b, weight=0.5):
	ax, ay = a[0], a[1]
	bx, by = b[0], b[1]
	aw, bw = 1.0 - weight, weight
	return (ax * aw + bx * bw, ay * aw + by * bw)

def rotateVector(vector, angle):
	x, y = vector
	c, s = cos(angle), sin(angle)
	return ((c * x) - (s * y), (s * x) + (c * y))

def findMinMax(iterable):
	minv = float('inf')
	maxv = float('-inf')
	for val in iterable:
		if val < minv:
			minv = val
		if val > maxv:
			maxv = val
	return (minv, maxv)

def radialToCartesian(point, radius, heading):
	angle = heading + (math.pi / 2.0)
	rx, ry = radius * cos(angle), radius * sin(angle)
	return (point[0] + rx, point[1] + ry)

def positionRelativeToPoint(point, heading, offset):
	ro = rotateVector(offset, heading)
	return addVectors(point, ro)

def headingOfSegment(pointA, pointB):
	ax, ay = pointA
	bx, by = pointB
	return math.atan2(by - ay, bx - ax) - (math.pi / 2.0)

def viewAngleToPoint(point, base, heading):
	x, y = base
	ox, oy = point
	a = math.atan2(oy - y, ox - x) - (heading + (math.pi / 2.0))
	if a < -math.pi:
		a += math.tau
	elif a > math.pi:
		a -= math.tau
	assert -math.pi <= a and a <= math.pi
	return a

def apparentHeadingAtPoint(point, heading, base):
	x, y = base
	ox, oy = point
	a = (heading + (math.pi / 2.0)) - math.atan2(oy - y, ox - x)
	if a < -math.pi:
		a += math.tau
	elif a > math.pi:
		a -= math.tau
	assert -math.pi <= a and a <= math.pi
	return a

def circumcircleOfAnnulus(center, heading, angle, minDist, maxDist):
	m = (minDist + maxDist) / 2.0
	g = (maxDist - minDist) / 2.0
	h = m * math.sin(angle / 2.0)
	h2 = h * h
	d = math.sqrt(h2 + (m * m))
	r = math.sqrt(h2 + (g * g))
	return radialToCartesian(center, d, heading), r

def pointIsInCone(point, base, heading, angle):
	va = viewAngleToPoint(point, base, heading)
	return (abs(va) <= angle / 2.0)

def polygonUnion(polys, tolerance=0.05, holeTolerance=0.002):
	union = shapely.ops.unary_union(list(polys))
	assert union.is_valid, union
	if tolerance > 0:
		if isinstance(union, shapely.geometry.MultiPolygon):
			polys = [cleanPolygon(poly, tolerance, holeTolerance) for poly in union]
			union = shapely.ops.unary_union(polys)
		else:
			union = cleanPolygon(union, tolerance, holeTolerance)
	assert union.is_valid, union
	return union

def cleanPolygon(poly, tolerance, holeTolerance):
	exterior = cleanChain(poly.exterior.coords, tolerance)
	assert len(exterior) >= 3
	ints = []
	for interior in poly.interiors:
		interior = cleanChain(interior.coords, tolerance)
		if len(interior) >= 3:
			hole = shapely.geometry.Polygon(interior)
			if hole.area > holeTolerance:
				ints.append(interior)
	newPoly = shapely.geometry.Polygon(exterior, ints)
	assert newPoly.is_valid, newPoly
	return newPoly

def cleanChain(chain, tolerance, angleTolerance=0.008):
	if len(chain) <= 2:
		return chain
	closed = (tuple(chain[0]) == tuple(chain[-1]))
	tol2 = tolerance * tolerance
	# collapse hooks
	chain = np.array(chain)
	a, b = chain[-1], chain[0]
	newChain = []
	for c in chain[1:]:
		dx, dy = c[0] - a[0], c[1] - a[1]
		if (dx * dx) + (dy * dy) > tol2:
			newChain.append(b)
			a = b
			b = c
		else:
			b = c
	if len(newChain) == 0:
		return newChain
	newChain.append(newChain[0] if closed else c)
	return newChain

#: Whether to warn when falling back to pypoly2tri for triangulation
givePP2TWarning = True

class TriangulationError(RuntimeError):
	"""Signals that the installed triangulation libraries are insufficient.

	Specifically, raised when pypoly2tri hits the recursion limit trying to
	triangulate a large polygon.
	"""
	pass

def triangulatePolygon(polygon):
	"""Triangulate the given Shapely polygon.

	Note that we can't use ``shapely.ops.triangulate`` since it triangulates
	point sets, not polygons (i.e., it doesn't respect edges). We need an
	algorithm for triangulation of polygons with holes (it doesn't need to be a
	Delaunay triangulation).

	We currently use the GPC library (wrapped by the ``Polygon3`` package) if it is
	installed. Since it is not free for commercial use, we don't require it as a
	dependency, falling back on the BSD-compatible ``pypoly2tri`` as needed. In this
	case we issue a warning, since GPC is more robust and handles large polygons.
	The warning can be disabled by setting `givePP2TWarning` to ``False``.

	Args:
		polygon (shapely.geometry.Polygon): Polygon to triangulate.

	Returns:
		A list of disjoint (except for edges) triangles whose union is the
		original polygon.
	"""
	try:
		import Polygon
		return triangulatePolygon_gpc(polygon)
	except ImportError:
		pass
	if givePP2TWarning:
		warnings.warn('Using pypoly2tri for triangulation; for non-commercial use, consider'
		              ' installing the faster Polygon3 library (pip install Polygon3)')
	return triangulatePolygon_pypoly2tri(polygon)

def triangulatePolygon_pypoly2tri(polygon):
	polyline = []
	for c in polygon.exterior.coords[:-1]:
		polyline.append(pypoly2tri.shapes.Point(c[0],c[1]))
	cdt = pypoly2tri.cdt.CDT(polyline)
	for i in polygon.interiors:
		polyline = []
		for c in i.coords[:-1]:
			polyline.append(pypoly2tri.shapes.Point(c[0],c[1]))
		cdt.AddHole(polyline)
	try:
		cdt.Triangulate()
	except RecursionError:		# polygon too big for pypoly2tri
		raise TriangulationError('pypoly2tri unable to triangulate large polygon; for '
		                         'non-commercial use, try "pip install Polygon3"')

	triangles = list()
	for t in cdt.GetTriangles():
		triangles.append(shapely.geometry.Polygon([
			t.GetPoint(0).toTuple(),
			t.GetPoint(1).toTuple(),
			t.GetPoint(2).toTuple()
		]))
	return triangles

def triangulatePolygon_gpc(polygon):
	import Polygon as PolygonLib
	poly = PolygonLib.Polygon(polygon.exterior.coords)
	for interior in polygon.interiors:
		poly.addContour(interior.coords, True)
	tristrips = poly.triStrip()
	triangles = []
	for strip in tristrips:
		a, b = strip[:2]
		for c in strip[2:]:
			tri = shapely.geometry.Polygon((a, b, c))
			triangles.append(tri)
			a = b
			b = c
	return triangles

def plotPolygon(polygon, plt, style='r-'):
	def plotRing(ring):
		x, y = ring.xy
		plt.plot(x, y, style)
	if isinstance(polygon, shapely.geometry.MultiPolygon):
		polygons = polygon
	else:
		polygons = [polygon]
	for polygon in polygons:
		plotRing(polygon.exterior)
		for ring in polygon.interiors:
			plotRing(ring)

class RotatedRectangle:
	"""mixin providing collision detection for rectangular objects and regions"""
	def containsPoint(self, point):
		diff = point - self.position.toVector()
		x, y = diff.rotatedBy(-self.heading)
		return abs(x) <= self.hw and abs(y) <= self.hh

	def intersects(self, rect):
		if not isinstance(rect, RotatedRectangle):
			raise RuntimeError(f'tried to intersect RotatedRectangle with {type(rect)}')
		# Quick check by bounding circles
		dx, dy = rect.position.toVector() - self.position.toVector()
		rr = self.radius + rect.radius
		if (dx * dx) + (dy * dy) > (rr * rr):
			return False
		# Check for separating line parallel to our edges
		if self.edgeSeparates(self, rect):
			return False
		# Check for separating line parallel to rect's edges
		if self.edgeSeparates(rect, self):
			return False
		return True

	@staticmethod
	def edgeSeparates(rectA, rectB):
		"""Whether an edge of rectA separates it from rectB"""
		base = rectA.position.toVector()
		rot = -rectA.heading
		rc = [(corner - base).rotatedBy(rot) for corner in rectB.corners]
		x, y = zip(*rc)
		minx, maxx = findMinMax(x)
		miny, maxy = findMinMax(y)
		if maxx < -rectA.hw or rectA.hw < minx:
			return True
		if maxy < -rectA.hh or rectA.hh < miny:
			return True
		return False

	@property
	def polygon(self):
		if any(needsSampling(c) or needsLazyEvaluation(c) for c in self.corners):
			return None		# can only convert fixed Regions to Polygons
		return self.getConstantPolygon()

	@utils.cached
	def getConstantPolygon(self):
		assert not any(needsSampling(c) or needsLazyEvaluation(c) for c in self.corners)
		# TODO refactor???
		corners = [(x, y) for x, y in self.corners]		# convert Vectors to tuples
		return shapely.geometry.Polygon(corners)



"""Scenic world model for the CARLA driving simulator.

This model is designed to be used with the CARLA interface to the VerifAI toolkit.
See the `VerifAI repository <https://github.com/BerkeleyLearnVerify/VerifAI>`_ for
further documentation and examples.

The model currently supports vehicles, pedestrians, and props.
Vehicles have an ``agent`` parameter, which specifies the agent to be used to control
the vehicle.

In addition, the model uses several global parameters to control weather (descriptions
are from the CARLA Python API reference):

    * ``cloudiness`` (float):
      Weather cloudiness. It only affects the RGB camera sensor. Values range from 0 to 100.
    * ``precipitation`` (float):
      Precipitation amount for controlling rain intensity. It only affects the RGB camera sensor. Values range from 0 to 100.
    * ``precipitation_deposits`` (float):
      Precipitation deposits for controlling the area of puddles on roads. It only affects the RGB camera sensor. Values range from 0 to 100.
    * ``wind_intensity`` (float):
      Wind intensity, it affects the clouds moving speed, the raindrop direction, and vegetation. This doesn't affect the car physics. Values range from 0 to 100.
    * ``sun_azimuth_angle`` (float):
      The azimuth angle of the sun in degrees. Values range from 0 to 360 (degrees).
    * ``sun_altitude_angle`` (float):
      Altitude angle of the sun in degrees. Values range from -90 to 90 (where 0 degrees is the horizon).

.. raw:: html

   <h2>Submodules</h2>

.. autosummary::
   :toctree: _autosummary

   model
   map
   interface
   car_models
   prop_models
"""



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['scenic',
 'scenic.core',
 'scenic.simulators',
 'scenic.simulators.carla',
 'scenic.simulators.formats',
 'scenic.simulators.formats.opendrive',
 'scenic.simulators.gta',
 'scenic.simulators.webots',
 'scenic.simulators.webots.guideways',
 'scenic.simulators.webots.mars',
 'scenic.simulators.webots.road',
 'scenic.simulators.xplane',
 'scenic.syntax']

package_data = \
{'': ['*']}

install_requires = \
['antlr4-python3-runtime>=4.8,<4.9',
 'dotmap>=1.3.13,<2.0.0',
 'matplotlib>=3.2.1,<4.0.0',
 'numpy>=1.18.2,<2.0.0',
 'opencv-python>=4.2.0,<5.0.0',
 'pillow>=7.1.1,<8.0.0',
 'pynverse>=0.1.4,<0.2.0',
 'pypoly2tri>=0.0.3,<0.0.4',
 'scipy>=1.4.1,<2.0.0',
 'shapely>=1.7.0,<2.0.0']

extras_require = \
{u'dev': ['pytest-randomly>=3.2.1,<4.0.0',
          'tox>=3.14.0,<4.0.0',
          'sphinx>=3.0.0,<4.0.0',
          'pytest>=5.4.1,<6.0.0',
          'pyproj>=2.6.0,<3.0.0',
          'sphinx_rtd_theme>=0.4.3,<0.5.0'],
 u'gpc_triangulation': ['Polygon3>=3.0.8,<4.0.0'],
 u'guideways': ['pyproj>=2.6.0,<3.0.0']}

entry_points = \
{u'pygments.lexers': ['scenic = scenic.syntax.pygment:ScenicLexer']}

setup_kwargs = {
    'name': 'scenic',
    'version': '1.1.0',
    'description': 'The Scenic scenario description language.',
    'long_description': '# Scenic\n\n[![Documentation Status](https://readthedocs.org/projects/scenic-lang/badge/?version=latest)](https://scenic-lang.readthedocs.io/en/latest/?badge=latest)\n[![License](https://img.shields.io/badge/License-BSD%203--Clause-blue.svg)](https://opensource.org/licenses/BSD-3-Clause)\n\nA compiler and scene generator for the Scenic scenario description language.\nPlease see the [documentation](https://scenic-lang.readthedocs.io/) for installation instructions, as well as tutorials and other information about the Scenic language, its implementation, and its interfaces to various simulators.\n\nFor a description of the language and some of its applications, see [our PLDI 2019 paper](https://arxiv.org/abs/1809.09310).\nScenic was designed and implemented by Daniel J. Fremont, Tommaso Dreossi, Shromona Ghosh, Xiangyu Yue, Alberto L. Sangiovanni-Vincentelli, and Sanjit A. Seshia.\n\nIf you have any problems using Scenic, please submit an issue to [our GitHub repository](https://github.com/BerkeleyLearnVerify/Scenic) or contact Daniel at <dfremont@ucsc.edu>.\n\nThe repository is organized as follows:\n\n* the _src/scenic_ directory contains the package proper;\n* the _examples_ directory has many examples of Scenic programs;\n* the _docs_ directory contains the sources for the documentation;\n* the _tests_ directory contains tests for the Scenic compiler.\n',
    'author': 'Daniel Fremont',
    'author_email': 'dfremont@ucsc.edu',
    'maintainer': 'Daniel Fremont',
    'maintainer_email': 'dfremont@ucsc.edu',
    'url': 'https://github.com/BerkeleyLearnVerify/Scenic',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""World model for road intersection scenarios in Webots.

This is a more specialized version of the :doc:`scenic.simulators.webots.road` model which also includes guideway information from the `Intelligent Intersections Toolkit <https://github.com/ucbtrans/intelligent_intersection>`_.

.. raw:: html

   <h2>Submodules</h2>

.. autosummary::
   :toctree: _autosummary

   model
   intersection
   interface
"""




"""Specifiers and associated objects."""

import itertools

from scenic.core.lazy_eval import (DelayedArgument, toDelayedArgument, requiredProperties,
                                   needsLazyEvaluation)
from scenic.core.distributions import toDistribution
from scenic.core.utils import RuntimeParseError

## Specifiers themselves

class Specifier:
	"""Specifier providing a value for a property given dependencies.

	Any optionally-specified properties are evaluated as attributes of the primary value.
	"""
	def __init__(self, prop, value, deps=None, optionals={}):
		self.property = prop
		self.value = toDelayedArgument(value)
		if deps is None:
			deps = set()
		deps |= requiredProperties(value)
		if prop in deps:
			raise RuntimeParseError(f'specifier for property {prop} depends on itself')
		self.requiredProperties = deps
		self.optionals = optionals

	def applyTo(self, obj, optionals):
		"""Apply specifier to an object, including the specified optional properties."""
		val = self.value.evaluateIn(obj)
		val = toDistribution(val)
		assert not needsLazyEvaluation(val)
		setattr(obj, self.property, val)
		for opt in optionals:
			assert opt in self.optionals
			setattr(obj, opt, getattr(val, opt))

	def __str__(self):
		return f'<Specifier of {self.property}>'

## Support for property defaults

class PropertyDefault:
	"""A default value, possibly with dependencies."""
	def __init__(self, requiredProperties, attributes, value):
		self.requiredProperties = requiredProperties
		self.value = value

		def enabled(thing, default):
			if thing in attributes:
				attributes.remove(thing)
				return True
			else:
				return default
		self.isAdditive = enabled('additive', False)
		for attr in attributes:
			raise RuntimeParseError(f'unknown property attribute "{attr}"')

	@staticmethod
	def forValue(value):
		if isinstance(value, PropertyDefault):
			return value
		else:
			return PropertyDefault(set(), set(), lambda self: value)

	def resolveFor(self, prop, overriddenDefs):
		"""Create a Specifier for a property from this default and any superclass defaults."""
		if self.isAdditive:
			allReqs = self.requiredProperties
			for other in overriddenDefs:
				allReqs |= other.requiredProperties
			def concatenator(context):
				allVals = [self.value(context)]
				for other in overriddenDefs:
					allVals.append(other.value(context))
				return tuple(allVals)
			val = DelayedArgument(allReqs, concatenator)
		else:
			val = DelayedArgument(self.requiredProperties, self.value)
		return Specifier(prop, val)



# Generated from WBT.g4 by ANTLR 4.8
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .WBTParser import WBTParser
else:
    from WBTParser import WBTParser

# This class defines a complete generic visitor for a parse tree produced by WBTParser.

class WBTVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by WBTParser#world.
    def visitWorld(self, ctx:WBTParser.WorldContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WBTParser#defn.
    def visitDefn(self, ctx:WBTParser.DefnContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WBTParser#node.
    def visitNode(self, ctx:WBTParser.NodeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WBTParser#nodeBody.
    def visitNodeBody(self, ctx:WBTParser.NodeBodyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WBTParser#attribute.
    def visitAttribute(self, ctx:WBTParser.AttributeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WBTParser#value.
    def visitValue(self, ctx:WBTParser.ValueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WBTParser#vector.
    def visitVector(self, ctx:WBTParser.VectorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WBTParser#string.
    def visitString(self, ctx:WBTParser.StringContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WBTParser#array.
    def visitArray(self, ctx:WBTParser.ArrayContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WBTParser#vectorWithNewlines.
    def visitVectorWithNewlines(self, ctx:WBTParser.VectorWithNewlinesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by WBTParser#boolean.
    def visitBoolean(self, ctx:WBTParser.BooleanContext):
        return self.visitChildren(ctx)



del WBTParser